import sys, re, json
from xml.etree import ElementTree

def insert_n(s, nb):
    sout = ""
    def sub(g):
        if g.group(2):
            a, b = int(g.group(1)), int(g.group(2)[1:])
            return nb[-a - 1:-b or None]
        else:
            a = int(g.group(1))
            return nb[-a - 1]

    s = re.sub(r'n\[(\d+)(:\d+)?\]', sub, s)
    s = "".join(s.split(":"))
    return int(s.replace("0b", ""), 2)

def parse_one(regs, xml):
    t = ElementTree.parse(xml)

    for reg in t.findall('registers/register'):
        data = {}

        name = reg.find('reg_short_name').text
        fullname = reg.find('reg_long_name').text

        if name.startswith("S3_") or name.startswith("SYS S1_"):
            continue

        array = reg.find('reg_array')

        start = end = 0

        if array:
            start = int(array.find("reg_array_start").text)
            end = int(array.find("reg_array_end").text)

        encs = {}
        accessors = {}

        for am in reg.findall('access_mechanisms/access_mechanism'):
            accessor = am.attrib["accessor"]
            if accessor.startswith("MSRimmediate"):
                continue
            ins = am.find("encoding/access_instruction").text.split(" ")[0]
            regname = accessor.split(" ", 1)[1]
            enc = {}
            for e in am.findall("encoding/enc"):
                enc[e.attrib["n"]] = e.attrib["v"]

            enc = enc["op0"], enc["op1"], enc["CRn"], enc["CRm"], enc["op2"]
            if regname in encs:
                assert encs[regname] == enc
            encs[regname] = enc
            accessors.setdefault(regname, set()).add(ins)

        if not encs:
            continue

        fieldsets = []

        width = None

        for fields_elem in reg.findall('reg_fieldsets/fields'):

            fieldset = {}

            if (instance_elem := fields_elem.find('fields_instance')) is not None:
                fieldset["instance"] = instance_elem.text

            fields = []

            set_width = int(fields_elem.attrib["length"])

            if width is None:
                width = set_width
            else:
                assert width == set_width

            single_field = False

            for f in fields_elem.findall('field'):

                if f.attrib.get("rwtype", None) in ("RES0", "RES1", "RAZ", "RAZ/WI", "RAO/WI", "UNKNOWN"):
                    continue
                msb, lsb = int(f.find('field_msb').text), int(f.find('field_lsb').text)

                assert not single_field

                if msb == width - 1 and lsb == 0:
                    continue

                if (name_elem := f.find('field_name')) is not None:
                    name = name_elem.text
                else:
                    assert not fields
                    continue

                field = {
                    "name": name,
                    "msb": msb,
                    "lsb": lsb,
                }
                fields.append(field)

            fields.sort(key=lambda x: x["lsb"], reverse=True)

            fieldset["fields"] = fields
            fieldsets.append(fieldset)

        for idx, n in enumerate(range(start, end + 1)):
            nb = "{0:064b}".format(n)[::-1]
            for name, enc in sorted(encs.items()):
                enc = tuple(insert_n(i, nb) for i in enc)
                data = {
                    "index": idx,
                    "name": name.replace("<n>", "%d" % n),
                    "fullname": fullname,
                    "enc": enc,
                    "accessors": sorted(list(accessors[name])),
                    "fieldsets": fieldsets,
                }

                if width is not None:
                    data["width"] = width

                yield data

if __name__ == "__main__":
    regs = []
    for i in sys.argv[1:]:
        regs.extend(parse_one(regs, i))

    json.dump(regs, sys.stdout)






"""
I think these are executed by a simple state machine on the firmware's arm core,
and the typical result is a commandlist submitting to one of the gpu's hardware
command processors.

It seems like a common pattern is:
  1. Start (3D or Compute)
  2. Timestamp
  3. Wait For Interrupts
  4. Timestamp again
  5. Finish (3D or Compute)
  6. End

Error messages call these as SKU commands

"""
from m1n1.constructutils import *

from construct import *
from construct.core import Int64ul, Int32ul, Int32sl
import textwrap

__all__ = []

class TimeStamp(ConstructValueClass):
    subcon = Int64ul

    def __init__(self, value=0):
        self.value = value

class TimeStampPointers(ConstructClass):
    subcon = Struct(
        "start_addr" / Int64ul,
        "end_addr" / Int64ul,
    )

    def __init__(self, start_addr=0, end_addr=0):
        super().__init__()
        self.start_addr = start_addr
        self.end_addr = end_addr

class TsFlag(ConstructValueClass):
    subcon = Int8ul

    def __init__(self, value=0):
        self.value = value

class WrappedPointer(ConstructValueClass):
    subcon = Int64ul

    def __init__(self, value=0):
        self.value = value

class StampCounter(ConstructValueClass):
    subcon = Hex(Int32ul)

    def __init__(self):
        self.value = 0

class BufferManagerBlockControl(ConstructClass):
    subcon = Struct(
        "total" / Int32ul,
        "wptr" / Int32ul,
        "unk" / Int32ul,
        "pad" / ZPadding(0x34)
    )

class BufferManagerCounter(ConstructClass):
    subcon = Struct(
        "count" / Int32ul,
        "pad" / ZPadding(0x3c)
    )

class BufferManagerMisc(ConstructClass):
    subcon = Struct(
        "gpu_0" / Default(Int32ul, 0),
        "gpu_4" / Default(Int32ul, 0),
        "gpu_8" / Default(Int32ul, 0),
        "gpu_c" / Default(Int32ul, 0),
        "pad_10" / ZPadding(0x10),
        "cpu_flag" / Int32ul,
        "pad_24" / ZPadding(0x1c),
    )

class BufferManagerInfo(ConstructClass):
    subcon = Struct(
        "gpu_counter" / Int32ul,
        "unk_4" / Int32ul,
        "last_id" / Int32ul,
        "cur_id" / Int32ul,
        "unk_10" / Int32ul,
        "gpu_counter2" / Int32ul,
        "unk_18" / Int32ul,
        Ver("V < V13_0B4 || G >= G14X", "unk_1c" / Int32ul),
        "page_list_addr" / Int64ul,
        "page_list_size" / Int32ul,
        "page_count" / Int32ul,
        "unk_30" / Int32ul,
        "block_count" / Int32ul,
        "unk_38" / Int32ul,
        "block_list_addr" / Int64ul,
        "block_ctl_addr" / Int64ul, # points to two u32s
        "block_ctl" / ROPointer(this.block_ctl_addr, BufferManagerBlockControl),
        "last_page" / Int32ul,
        "gpu_page_ptr1" / Int32ul,
        "gpu_page_ptr2" / Int32ul,
        "unk_58" / Int32ul,
        "block_size" / Int32ul,
        "unk_60" / Int64ul,
        "counter_addr" / Int64ul,
        "counter" / ROPointer(this.counter_addr, BufferManagerCounter),
        "unk_70" / Int32ul,
        "unk_74" / Int32ul,
        "unk_78" / Int32ul,
        "unk_7c" / Int32ul,
        "unk_80" / Int32ul,
        "unk_84" / Int32ul,
        "unk_88" / Int32ul,
        "unk_8c" / Int32ul,
        "unk_90" / HexDump(Bytes(0x30)),

    )

    def __init__(self):
        super().__init__()
        self.gpu_counter = 0x0
        self.unk_4 = 0
        self.last_id = 0x0
        self.cur_id = 0xffffffff
        self.unk_10 = 0x0
        self.gpu_counter2 = 0x0
        self.unk_18 = 0x0
        self.unk_1c = 0x0
        self.unk_30 = 0xd1a
        self.unk_38 = 0x0
        self.gpu_page_ptr1 = 0x0
        self.gpu_page_ptr2 = 0x0
        self.unk_58 = 0x0
        self.unk_60 = 0x0
        self.unk_70 = 0x0
        self.unk_74 = 0x0
        self.unk_78 = 0x0
        self.unk_7c = 0x0
        self.unk_80 = 0x1
        self.unk_84 = 0x66cc
        self.unk_88 = 0x2244
        self.unk_8c = 0x0
        self.unk_90 = bytes(0x30)

class RegisterDefinition(ConstructClass):
    subcon = Struct(
        "number" / Int32ul,
        "data" / Int64ul,
    )

    def __init__(self, number, data):
        super().__init__()
        self.number = number
        self.data = data

class Start3DClearPipelineBinding(ConstructClass):
    subcon = Struct(
        "pipeline_bind" / Int64ul,
        "address" / Int64ul,
    )

    def __init__(self, pipeline_bind=None, address=None):
        super().__init__()
        self.pipeline_bind = pipeline_bind
        self.address = address

class Start3DStorePipelineBinding(ConstructClass):
    subcon = Struct(
        "unk_0" / Int64ul,
        "unk_8" / Int32ul,
        "pipeline_bind" / Int32ul,
        "unk_10" / Int32ul,
        "address" / Int32ul,
        "unk_18" / Int32ul,
        "unk_1c_padding" / Int32ul,
    )

    def __init__(self, pipeline_bind=None, address=None):
        super().__init__()
        self.unk_0 = 0
        self.unk_8 = 0
        self.pipeline_bind = pipeline_bind
        self.unk_10 = 0
        self.address = address
        self.unk_18 = 0
        self.unk_1c_padding = 0

class Start3DArrayAddr(ConstructClass):
    subcon = Struct(
        "ptr" / Int64ul,
        "unk_padding" / Int64ul,
    )

    def __init__(self, ptr=None):
        super().__init__()
        self.ptr = ptr
        self.unk_padding = 0

class AuxFBInfo(ConstructClass):
    subcon = Struct(
        "unk1" / Int32ul,
        "unk2" / Int32ul,
        "width" / Dec(Int32ul),
        "height" / Dec(Int32ul),
        Ver("V >= V13_0B4", "unk3" / Int64ul),
    )

    def __init__(self, unk1, unk2, width, height):
        super().__init__()
        self.unk1 = unk1
        self.unk2 = unk2
        self.width = width
        self.height = height
        self.unk3 = 0x100000

class Start3DStruct1(ConstructClass):
    subcon = Struct(
        "store_pipeline_bind" / Int32ul, # 0x12, 0x34 seen
        "store_pipeline_addr" / Int32ul,
        "unk_8" / Int32ul,
        "unk_c" / Int32ul,
        "merge_upper_x" / Float32l,
        "merge_upper_y" / Float32l,
        "unk_18" / Int64ul,
        "tile_blocks_y" / Int16ul, # * 4
        "tile_blocks_x" / Int16ul, # * 4
        "unk_24" / Int32ul,
        "tile_counts" / Int32ul,
        "unk_2c" / Int32ul,
        "depth_clear_val1" / Float32l,
        "stencil_clear_val1" / Int8ul,
        "unk_35" / Int8ul,
        "unk_36" / Int16ul,
        "unk_38" / Int32ul,
        "unk_3c" / Int32ul,
        "helper_cfg" / Int32ul,
        "unk_44_padding" / HexDump(Bytes(0x9c)),
    )

class Start3DStruct2(ConstructClass):
    subcon = Struct(
        "unk_0" / Int64ul,
        "clear_pipeline" / Start3DClearPipelineBinding,
        "unk_18" / Int64ul,
        "scissor_array" / Int64ul,
        "depth_bias_array" / Int64ul,
        "aux_fb" / AuxFBInfo,
        "depth_dimensions" / Int64ul,
        "visibility_result_buffer" / Int64ul,
        "depth_flags" / Int64ul, # 0x40000 - has stencil 0x80000 - has depth
        Ver("G >= G14", "unk_58_g14_0" / Int64ul),
        Ver("G >= G14", "unk_58_g14_8" / Int64ul),
        "depth_buffer_ptr1" / Int64ul,
        "depth_buffer_ptr2" / Int64ul,
        "stencil_buffer_ptr1" / Int64ul,
        "stencil_buffer_ptr2" / Int64ul,
        Ver("G >= G14", "unk_68_g14_0" / HexDump(Bytes(0x20))),
        "unk_78" / Array(4, Int64ul),
        "depth_aux_buffer_ptr1" / Int64ul,
        "unk_a0" / Int64ul,
        "depth_aux_buffer_ptr2" / Int64ul,
        "unk_b0" / Int64ul,
        "stencil_aux_buffer_ptr1" / Int64ul,
        "unk_c0" / Int64ul,
        "stencil_aux_buffer_ptr2" / Int64ul,
        "unk_d0" / Int64ul,
        "tvb_tilemap" / Int64ul,
        "tvb_heapmeta_addr" / Int64ul,
        "unk_e8" / Int64ul,
        "tvb_heapmeta_addr2" / Int64ul,
        "unk_f8" / Int64ul,
        "aux_fb_ptr" / Int64ul,
        "unk_108" / Array(6, Int64ul),
        "pipeline_base" / Int64ul,
        "unk_140" / Int64ul,
        "helper_program" / Int32ul, # bit 0: enable
        "unk_14c" / Int32ul, # 0
        "helper_arg" / Int64ul, # stack layout
        "unk_158" / Int64ul,
        "unk_160" / Int64ul,
        Ver("G < G14", "unk_168_padding" / HexDump(Bytes(0x1d8))),
        Ver("G >= G14", "unk_198_padding" / HexDump(Bytes(0x1a8))),
        Ver("V < V13_0B4", ZPadding(8)),
    )

class Start3DStruct3(ConstructClass):
    subcon = Struct(
        "registers_addr" / Int64ul,
        "register_count" / Int16ul,
        "registers_length" / Int16ul,
        "unk_d8" / Int32ul,
        "depth_bias_array" / Start3DArrayAddr,
        "scissor_array" / Start3DArrayAddr,
        "visibility_result_buffer" / Int64ul,
        "unk_118" / Int64ul,
        "unk_120" / Array(37, Int64ul),
        "unk_reload_pipeline" / Start3DClearPipelineBinding,
        "unk_258" / Int64ul,
        "unk_260" / Int64ul,
        "unk_268" / Int64ul,
        "unk_270" / Int64ul,
        "reload_pipeline" / Start3DClearPipelineBinding,
        "depth_flags" / Int64ul, # 0x40000 - has stencil 0x80000 - has depth
        "unk_290" / Int64ul,
        "depth_buffer_ptr1" / Int64ul,
        "unk_2a0" / Int64ul,
        "unk_2a8" / Int64ul,
        "depth_buffer_ptr2" / Int64ul,
        "depth_buffer_ptr3" / Int64ul,
        "depth_aux_buffer_ptr" / Int64ul,
        "stencil_buffer_ptr1" / Int64ul,
        "unk_2d0" / Int64ul,
        "unk_2d8" / Int64ul,
        "stencil_buffer_ptr2" / Int64ul,
        "stencil_buffer_ptr3" / Int64ul,
        "stencil_aux_buffer_ptr" / Int64ul,
        "unk_2f8" / Array(2, Int64ul),
        "aux_fb_unk0" / Int32ul,
        "unk_30c" / Int32ul,
        "aux_fb" / AuxFBInfo,
        "s2_unk_f8" / Int32ul,
        "unk_324_padding" / HexDump(Bytes(0xc)),
        "unk_partial_store_pipeline" / Start3DStorePipelineBinding,
        "partial_store_pipeline" / Start3DStorePipelineBinding,
        "depth_clear_val2" / Float32l,
        "stencil_clear_val2" / Int8ul,
        "unk_375" / Int8ul,
        "unk_376" / Int16ul,
        "unk_378" / Int32ul,
        "unk_37c" / Int32ul,
        "unk_380" / Int64ul,
        "unk_388" / Int64ul,
        Ver("V >= V13_0B4", "unk_390_0" / Int64ul),
        "depth_dimensions" / Int64ul,
    )

class BufferThing(ConstructClass):
    subcon = Struct(
        Ver("G >= G14X", "unk0_addr" / Int64ul),
        Ver("G >= G14X", "unk0_addr2" / Int64ul),
        # Ver("G >= G14X", "unk0" / ROPointer(this.unk0_addr, Array(8, Int32ul))),
        "unk_0" / Int64ul,
        "unk_8" / Int64ul,
        "unk_10" / Int64ul,
        "unkptr_18" / Int64ul,
        "unk_20" / Int32ul,
        Ver("V >= V13_3", "unk_28" / Int64ul),
        "bm_misc_addr" / Int64ul,
        "bm_misc" / ROPointer(this.bm_misc_addr, BufferManagerMisc),
        "unk_2c" / Int32ul,
        Ver("G < G14X", "unk_30" / Int64ul),
        Ver("G < G14X", "unk_38" / Int64ul),
    )

class Start3DStruct6(ConstructClass):
    subcon = Struct(
        "tvb_overflow_count" / Int64ul,
        "unk_8" / Int32ul,
        "unk_c" / Int32ul,
        "unk_10" / Int32ul,
        "encoder_id" / Int64ul,
        "unk_1c" / Int32ul,
        "sampler_array" / Int64ul,
        "sampler_count" / Int32ul,
        "sampler_max" / Int32ul,
        "unk_30" / Int32ul,
        "unk_34" / Int32ul,
        "unk_38" / Int32ul,
    )

class Start3DStruct7(ConstructClass):
    subcon = Struct(
        Ver("V >= V13_3", "unk_0_0" / Int32ul),
        "unk_0" / Int64ul,
        "stamp1_addr" / WrappedPointer, # same contents as below
        "stamp1" / ROPointer(this.stamp1_addr.value, StampCounter),
        "stamp2_addr" / WrappedPointer, # same as FinalizeComputeCmd.stamp - some kind of fence/token
        "stamp2" / ROPointer(this.stamp2_addr.value, StampCounter),
        "stamp_value" / Int32ul,
        "ev_3d" / Int32ul,
        "evctl_index" / Int32ul,
        "unk_24" / Int32ul,
        "uuid" / Int32ul,
        "queue_cmd_count" / Int32ul,
        "unk_30" / Int32ul,
    )

    def __init__(self):
        super().__init__()
        self.stamp1_addr = StampCounter()
        self.stamp2_addr = StampCounter()

class Attachment(ConstructClass):
    subcon = Struct(
        "address" / Int64ul,
        "size" / Int32ul,
        "unk_c" / Int16ul,
        "unk_e" / Int16ul,
    )

    def __init__(self, addr, size, unk_c, unk_e):
        self.address = addr
        self.size = size
        self.unk_c = unk_c
        self.unk_e = unk_e

class Start3DCmd(ConstructClass):
    subcon = Struct( # 0x194 bytes''''
        "magic" / Const(0x24, Int32ul),
        "struct1_addr" / Int64ul, # empty before run. Output? WorkCommand3D + 0x3c0
        "struct2_addr" / Int64ul, # ??  WorkCommand3D + 0x78
        Ver("G >= G14X", "registers_addr" / Int64ul),
        "buf_thing_addr" / Int64ul,
        "buf_thing" / ROPointer(this.buf_thing_addr, BufferThing),
        "stats_ptr" / Int64ul,
        "busy_flag_ptr" / Int64ul, # 4 bytes
        "struct6_addr" / Int64ul, # 0x3c bytes
        "struct6" / ROPointer(this.struct6_addr, Start3DStruct6),
        "struct7_addr" / Int64ul, # 0x34 bytes
        "struct7" / ROPointer(this.struct7_addr, Start3DStruct7),
        "cmdqueue_ptr" / Int64ul, # points back to the CommandQueueInfo that this command came from
        "workitem_ptr" / Int64ul, # points back at the WorkItem that this command came from
        "context_id" / Int32ul,
        "unk_50" / Int32ul,
        "submission_id" / Int32ul,
        "buffer_mgr_slot" / Int32ul,
        "unk_5c" / Int32ul,
        "queue_cmd_count" / Int64ul, # 0
        "unk_68" / Int32ul, # 0
        "unk_buf_ptr" / Int64ul,
        "unk_buf2_ptr" / Int64ul, # 0x18 bytes
        Ver("V >= V13_3", "unk_7c_0" / Int64ul),
        "unk_7c" / Int32ul,
        "unk_80" / Int32ul,
        "unk_84" / Int32ul,
        "uuid" / Int32ul, # uuid for tracking
        "attachments" / Array(16, Attachment),
        "num_attachments" / Int32ul,
        "unk_190" / Int32ul,
        Ver("V >= V13_0B4", "counter" / Int64ul),
        Ver("V >= V13_0B4", "unkptr_19c" / Int64ul),
    )


class Finalize3DCmd(ConstructClass):
    subcon = Struct( # 0x9c bytes
        "magic" / Const(0x25, Int32ul),
        "uuid" / Int32ul, # uuid for tracking
        "unk_8" / Int32ul, # 0
        "stamp_addr" / Int64ul,
        "stamp" / ROPointer(this.stamp_addr, StampCounter),
        "stamp_value" / Int32ul,
        "unk_18" / Int32ul,
        "buf_thing_addr" / Int64ul,
        "buf_thing" / ROPointer(this.buf_thing_addr, BufferThing),
        "buffer_mgr_addr" / Int64ul,
        "buffer_mgr" / ROPointer(this.buffer_mgr_addr, BufferManagerInfo),
        "unk_2c" / Int64ul, # 1
        "stats_ptr" / Int64ul,
        "struct7_addr" / Int64ul,
        "struct7" / ROPointer(this.struct7_addr, Start3DStruct7),
        "busy_flag_ptr" / Int64ul,
        "cmdqueue_ptr" / Int64ul,
        "workitem_ptr" / Int64ul,
        "unk_5c" / Int64ul,
        "unk_buf_ptr" / Int64ul, # Same as Start3DCmd.unkptr_6c
        Ver("V >= V13_3", "unk_6c_0" / Int64ul),
        "unk_6c" / Int64ul, # 0
        "unk_74" / Int64ul, # 0
        "unk_7c" / Int64ul, # 0
        "unk_84" / Int64ul, # 0
        "unk_8c" / Int64ul, # 0
        Ver("G == G14 && V < V13_0B4", "unk_8c_g14" / Int64ul),
        "restart_branch_offset" / Int32sl,
        "unk_98" / Int32ul, # 1
        Ver("V >= V13_0B4", "unk_9c" / HexDump(Bytes(0x10))),
    )

class TilingParameters(ConstructClass):
    subcon = Struct(
        "size1" / Int32ul,
        "unk_4" / Int32ul,
        "unk_8" / Int32ul,
        "x_max" / Dec(Int16ul),
        "y_max" / Dec(Int16ul),
        "tile_count" / Int32ul,
        "x_blocks" / Int32ul,
        "y_blocks" / Int32ul,
        "size2" / Int32ul,
        "size3" / Int32ul,
        "unk_24" / Int32ul,
        "unk_28" / Int32ul,
        "helper_cfg" / Int32ul,
    )

class StartTACmdStruct2(ConstructClass):
    subcon = Struct(
        "unk_0" / Hex(Int64ul),
        "unk_8" / Hex(Int32ul),
        "unk_c" / Hex(Int32ul),
        "tvb_tilemap" / Hex(Int64ul),
        Ver("G < G14", "tvb_cluster_tilemaps" / Hex(Int64ul)),
        "tpc" / Hex(Int64ul),
        "tvb_heapmeta_addr" / Hex(Int64ul), # like Start3DStruct2.tvb_end_addr with bit 63 set?
        "iogpu_unk_54" / Int32ul,
        "iogpu_unk_55" / Int32ul,
        "iogpu_unk_56" / Int64ul,
        Ver("G < G14", "tvb_cluster_meta1" / Int64ul),
        "unk_48" / Int64ul,
        "unk_50" / Int64ul,
        "tvb_heapmeta_addr2" / Int64ul,
        Ver("G < G14", "unk_60" / Int64ul),
        Ver("G < G14", "core_mask" / Int64ul),
        "iogpu_deflake_1" / Int64ul,
        "iogpu_deflake_2" / Int64ul,
        "unk_80" / Int64ul,
        "iogpu_deflake_3" / Int64ul, # bit 50 set
        "encoder_addr" / Int64ul,
        Ver("G < G14", "tvb_cluster_meta2" / Int64ul),
        Ver("G < G14", "tvb_cluster_meta3" / Int64ul),
        Ver("G < G14", "tiling_control" / Int64ul),
        "unk_b0" / Array(6, Hex(Int64ul)),
        "pipeline_base" / Int64ul,
        Ver("G < G14", "tvb_cluster_meta4" / Int64ul),
        Ver("G < G14", "unk_f0" / Int64ul),
        "unk_f8" / Int64ul,
        "helper_program" / Int32ul, # bit 0: enable
        "unk_104" / Int32ul, # 0
        "helper_arg" / Int64ul, # stack layout
        "unk_110" / Hex(Int64ul),
        "unk_118" / Int32ul,
        Ver("G >= G14", ZPadding(8 * 9)),
    )

class StartTACmdStruct3(ConstructClass):
    subcon = Struct(
        "unk_480" / Array(6, Int32ul),
        "unk_498" / Int64ul,
        "unk_4a0" / Int32ul,
        "iogpu_deflake_1" / Int64ul,
        "unk_4ac" / Int32ul,
        "unk_4b0" / Int64ul,
        "unk_4b8" / Int32ul,
        "unk_4bc" / Int64ul,
        "unk_4c4_padding" / HexDump(Bytes(0x48)),
        "unk_50c" / Int32ul,
        "unk_510" / Int64ul,
        "unk_518" / Int64ul,
        "unk_520" / Int64ul,
        "unk_528" / Int32ul,
        "unk_52c" / Int32ul,
        "unk_530" / Int32ul,
        "encoder_id" / Int32ul,
        "unk_538" / Int32ul,
        "unk_53c" / Int32ul,
        "sampler_array" / Int64ul,
        "sampler_count" / Int32ul,
        "sampler_max" / Int32ul,
        "unk_550" / Array(6, Int32ul),
        "stamp1_addr" / WrappedPointer, # same contents as below
        "stamp1" / ROPointer(this.stamp1_addr.value, StampCounter),
        "stamp2_addr" / WrappedPointer, # same as FinalizeComputeCmd.stamp - some kind of fence/token
        "stamp2" / ROPointer(this.stamp2_addr.value, StampCounter),
        "stamp_value" / Int32ul,
        "ev_ta" / Int32ul,
        "evctl_index" / Int32ul, # 0-3
        "unk_584" / Int32ul,
        "uuid2" / Int32ul,
        "queue_cmd_count" / Int32ul,
        "unk_590" / Int32ul,
    )

    def __init__(self):
        super().__init__()
        self.stamp1_addr = StampCounter()
        self.stamp2_addr = StampCounter()

class StartTACmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x22, Int32ul),
        "tiling_params_addr" / Int64ul,
        "struct2_addr" / Int64ul,
        Ver("G >= G14X", "registers_addr" / Int64ul),
        "buffer_mgr_addr" / Int64ul,
        "buffer_mgr" / ROPointer(this.buffer_mgr_addr, BufferManagerInfo),
        "buf_thing_addr" / Int64ul,
        "buf_thing" / ROPointer(this.buf_thing_addr, BufferThing),
        "stats_ptr" / Int64ul,
        "cmdqueue_ptr" / Int64ul,
        "context_id" / Int32ul,
        "unk_38" / Int32ul,
        "submission_id" / Int32ul,
        "buffer_mgr_slot" / Int64ul,
        "unk_48" / Int64ul,
        "unk_50" / Int32ul,
        "struct3_addr" / Int64ul,
        "unkptr_5c" / Int64ul,
        "unk_5c" / ROPointer(this.unkptr_5c, HexDump(Bytes(0x18))),
        "unk_64" / Int32ul,
        "unk_68" / Int32ul,
        "uuid" / Int32ul,
        "unk_70" / Int32ul,
        "unk_74" / Array(29, Int64ul),
        "unk_15c" / Int32ul,
        "unk_160" / Int64ul,
        "unk_168" / Int32ul,
        "unk_16c" / Int32ul,
        "unk_170" / Int64ul,
        Ver("V >= V13_0B4", "counter" / Int64ul),
        Ver("V >= V13_0B4", "unkptr_180" / Int64ul),
        "unk_178" / Int32ul,
    )

class FinalizeTACmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x23, Int32ul),
        "buf_thing_addr" / Int64ul,
        "buf_thing" / ROPointer(this.buf_thing_addr, BufferThing),
        "buffer_mgr_addr" / Int64ul,
        "buffer_mgr" / ROPointer(this.buffer_mgr_addr, BufferManagerInfo),
        "stats_ptr" / Int64ul,
        "cmdqueue_ptr" / Int64ul, #
        "context_id" / Int32ul,
        "unk_28" / Int32ul,
        "struct3_addr" / Int64ul,
        "unk_34" / Int32ul,
        "uuid" / Int32ul,
        "stamp_addr" / Int64ul,
        "stamp" / ROPointer(this.stamp_addr, StampCounter),
        "stamp_value" / Int32ul,
        "unk_48" / Int64ul,
        "unk_50" / Int32ul,
        "unk_54" / Int32ul,
        "unk_58" / Int64ul,
        "unk_60" / Int32ul,
        "unk_64" / Int32ul,
        "unk_68" / Int32ul,
        Ver("G == G14 && V < V13_0B4", "unk_6c_g14" / Int64ul),
        "restart_branch_offset" / Int32sl,
        "unk_70" / Int32ul,
        Ver("V >= V13_0B4", "unk_74" / HexDump(Bytes(0x10))),
    )

class ComputeArgs(ConstructClass):
    subcon = Struct(
        unk = Bytes(0x7fa0),
        arg_buffers = Array(8, Int64ul),
        threadgroups_per_grid_addr = Int64ul,
        threads_per_threadgroup_addr = Int64ul,
    )

class JobMeta(ConstructClass):
    subcon = Struct(
        "unk_0" / Int32ul,
        "stamp_addr" / WrappedPointer, # same contents as below
        "stamp" / ROPointer(this.stamp_addr.value, StampCounter),
        "fw_stamp_addr" / WrappedPointer, # same as FinalizeComputeCmd.stamp - some kind of fence/token
        "stamp" / ROPointer(this.fw_stamp_addr.value, StampCounter),
        "stamp_value" / Int32ul,
        "stamp_slot" / Int32ul,
        "evctl_index" / Int32ul,
        "unk_20" / Int32ul,
        "uuid" / Int32ul,
        "queue_cmd_count" / Int32ul,
    )

class EncoderParams(ConstructClass):
    subcon = Struct(
        "unk_0" / Int32ul,
        "unk_4" / Int32ul,
        "unk_8" / Int32ul,
        "encoder_id" / Int32ul,
        "unk_10" / Int32ul,
        "iogpu_compute_unk44" / Int32ul,
        "seq_buffer" / Int64ul,
        "unk_1c" / Int64ul,
    )

class SubBufferList(ConstructClass):
    subcon = Struct(
        # Least significant byte encoding
        # 7654 3210
        # ==== ====
        # 0000 0000 - invalid
        # aaaa aa01 - Block size 0x400 or continuation
        # aaaa 0011 - Block size 0x1000
        # aa00 0111 - Block size 0x4000
        # 0000 1111 - Block size 0x10000
        # 0001 1111 - Block size 0x40000
        # 0011 1111 - Block size 0x100000? (not seen)
        # 0111 1111 - Block size 0x400000
        "buffers" / Array(256 + 12, Array(4, Int32ul))
    )

class HelperArg(ConstructClass):
    subcon = Struct(
        "core_stride" / Int32ul,
        "alloc_size_map" / Array(16, Int16ul),
        "max_subgroups" / Array(16, Int16ul),
        "core_list" / Array(128, Int16ul),
    )

class ComputeInfo(ConstructClass):
    # Only the cmdlist and pipelinebase and cmdlist fields are strictly needed to launch a basic
    # compute shader.
    subcon = Struct( # 0x1c bytes
        "iogpu_deflake_1" / Int64ul, # ComputeArgs
        "encoder" / Int64ul, # CommandList from userspace
        "iogpu_deflake_2" / Int64ul, # size 8, null
        "iogpu_deflake_3" / Int64ul, # size 8, null
        "iogpu_deflake_4" / Int64ul, # size 8, null
        "iogpu_deflake_5" / Int64ul, # size 8, null
        "pipeline_base" / Int64ul, # 0x11_00000000: Used for certain "short" pointers like pipelines (and shaders?)
        "unk_38" / Int64ul, # always 0x8c60
        "helper_program" / Int32ul, # 0x98000; bit 0: enable
        "unk_44" / Int32ul, # 0
        "helper_arg" / Int64ul, # work layout
        "helper_cfg" / Int32ul, # 0x40 - Size? only if work layout is provided
        "unk_54" / Int32ul, # 0
        "unk_58" / Int32ul, # 1
        "unk_5c" / Int32ul, # 0
        "iogpu_unk_40" / Int32ul, # 0x1c
        "unk_pad" / ZPadding(0xec),
    )

# Related to "IOGPU Misc"
class ComputeInfo2(ConstructClass):
    subcon = Struct(
        Ver("V >= V13_0B4", "unk_0_0" / ZPadding(4)),
        "unk_0" / HexDump(Bytes(0x24)),
        "iogpu_deflake_1" / Int64ul,
        "encoder_end" / Int64ul,
        "unk_34" / HexDump(Bytes(0x20)),
        "unk_g14x" / Int32ul,
        "unk_58" / Int32ul,
        Ver("V < V13_0B4", "unk_5c" / ZPadding(4)),
    )

class StartComputeCmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x29, Int32ul),
        "unk_buf_addr" / Int64ul, # Pointer to WorkCommandCP.unk_buf
        "computeinfo_addr" / Int64ul,
        Ver("G >= G14X", "registers_addr" / Int64ul),
        "stats_ptr" / Int64ul,
        "cmdqueue_ptr" / Int64ul,
        "context_id" / Int32ul, # 4
        "unk_28" / Int32ul, # 1
        "counter1" / Int32ul,
        "counter2" / Int32ul,
        "unk_34" / Int32ul,
        "unk_38" / Int32ul,
        "computeinfo2_addr" / Int64ul,
        "computeinfo2" / ROPointer(this.computeinfo2_addr, ComputeInfo2),
        "unk_44" / Int32ul,
        "uuid" / Int32ul,
        "attachments" / Array(16, Attachment),
        "num_attachments" / Int32ul,
        "padding" / Bytes(4),
        Ver("V >= V13_0B4", "unk_flag_addr" / Int64ul),
        Ver("V >= V13_0B4", "counter" / Int64ul),
        Ver("V >= V13_0B4", "event_ctrl_buf_addr" / Int64ul),
    )

class FinalizeComputeCmd(ConstructClass):
    subcon = Struct( # 0x64 bytes''''
        "magic" / Const(0x2a, Int32ul),
        "unkptr_4" / Int64ul, # same as ComputeStartCmd.unkptr_14
        "cmdqueue_ptr" / Int64ul, # points back to the submitinfo
        "context_id" / Int32ul,
        Ver("V < V13_0B4", "unk_18" / Int32ul),
        "computeinfo2_addr" / Int64ul, # same as ComputeStartCmd.unkptr_3c
        "unk_24" / Int32ul,
        "uuid" / Int32ul,  # uuid for tracking?
        "stamp" / Int64ul,
        "stamp_value" / Int32ul, # Gets written to unkptr_2c (after stamp?)
        "unk_38" / Int32ul,
        "unk_3c" / Int32ul,
        "unk_40" / Int32ul,
        "unk_44" / Int32ul,
        "unk_48" / Int32ul,
        "unk_4c" / Int32ul,
        "unk_50" / Int32ul,
        "unk_54" / Int32ul,
        "unk_58" / Int32ul,
        Ver("G == G14 && V < V13_0B4", "unk_5c_g14" / Int64ul),
        "restart_branch_offset" / Int32sl, # relative offset from start of Finalize to StartComputeCmd
        "unk_60" / Int32ul,
        Ver("V >= V13_0B4", "unk_64" / HexDump(Bytes(0xd))),
        Ver("V >= V13_0B4", "unkptr_71" / Int64ul),
        Ver("V >= V13_0B4", "pad_79" / ZPadding(7)),
    )

class BlitInfo(ConstructClass):
    subcon = Struct(
        "unk_18" / Int64ul,
        "unk_20" / Int64ul,
        "unk_28" / Int64ul,
        "unk_30" / Int64ul,
        "unk_38" / Int64ul,
        "unk_40" / Int64ul,
        "unk_48" / HexDump(Bytes(0xa0)),
        "unkptr_e8" / Int64ul,
        "unk_f0" / Int64ul,
        "unkptr_f8" / Int64ul,
        "pipeline_base" / Int64ul,
        "unk_108" / Int64ul,
        "unk_110" / HexDump(Bytes(0x248)),
        "unk_358" / Int32ul,
        "unk_35c" / Int32ul,
        "unk_360" / Int32ul,
        "unk_364" / Int32ul,
        "unk_368" / Float32l,
        "unk_36c" / Float32l,
        "unk_370" / Int64ul,
        "unk_378" / Int64ul,
        "unk_380" / Int64ul,
        "unk_388" / Int64ul,
        "unk_390" / HexDump(Bytes(0xa8)),
    )

class BlitInfo2(ConstructClass):
    subcon = Struct(
        "unk_0" / HexDump(Bytes(0x24)),
        "unk_24" / Int64ul,
        "unk_470" / Int64ul,
        "unk_478" / Int32ul,
        "unk_47c" / Int32ul,
        "unk_480" / Int64ul,
        "unk_488" / Int64ul,
    )

class StartBlitCmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x26, Int32ul),
        "unkptr_4" / Int64ul,
        "unkptr_c" / Int64ul,
        "registers_addr" / Int64ul,
        Ver("G >= G14X", "unk_1c_0" / Int32ul),
        "unkptr_1c" / Int64ul,
        "unkptr_24" / Int64ul,
        "context_id" / Int32ul,
        "unk_30" / Int32ul,
        "submission_id" / Int32ul,
        "unk_38" / Int32ul,
        "unk_3c" / Int32ul,
        "unk_40" / Int32ul,
        "unkptr_44" / Int64ul,
        "unkptr_4c" / Int64ul,
        "uuid" / Int32ul,
        Ver("G < G14X", "unk_2c" / Int32ul),
        "attachments" / Array(16, Attachment),
        "num_attachments" / Int32ul,
        "unk_160" / Int32ul,
        Ver("V >= V13_0B4", "counter" / Int64ul),
        Ver("V >= V13_0B4", "unkptr_19c" / Int64ul),
    )

class FinalizeBlitCmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x27, Int32ul),
        "unkptr_4" / Int64ul,
        "unkptr_c" / Int64ul,
        "context_id" / Int32ul,
        "unk_18" / Int32ul,
        "unkptr_1c" / Int64ul,
        "uuid" / Int32ul,
        Ver("V < V13_3", "unk_28" / Int32ul),
        "stamp_addr" / Int64ul,
        "stamp" / ROPointer(this.stamp_addr, StampCounter),
        "stamp_value" / Int32ul,
        "unk_38" / HexDump(Bytes(0x24)),
        "restart_branch_offset" / Int32sl, # relative
        "unk_60" / Int32ul,
        Ver("V >= V13_0B4", "unk_5d8_0" / Int32ul),
        Ver("V >= V13_0B4", "unk_5d8_4" / Int8ul),
        Ver("V >= V13_0B4", "evctl_buf_addr" / Int64ul),
        Ver("V >= V13_0B4", "unk_5d8_4" / Array(3, Int8ul)),
        Ver("V >= V13_0B4", "unk_0" / Int32ul),
        Ver("V >= V13_0B4", "unkptr_1" / Int64ul),
        Ver("V >= V13_0B4", "unkptr_2" / Int64ul),
        Ver("V >= V13_0B4", "unkptr_3" / Int64ul),
        Ver("V >= V13_0B4", "unkptr_4" / Int64ul),
        Ver("V >= V13_0B4", "unkptr_5" / Int64ul),
        Ver("V >= V13_0B4", "unk_9c" / Int64ul),
        Ver("V >= V13_0B4", "unkptr_6" / Int64ul),
        Ver("V >= V13_0B4", "unk_ac" / HexDump(Bytes(0x30))),
        Ver("V >= V13_0B4", "unk_dc" / Int32ul),
        Ver("V >= V13_0B4", "unk_e0" / HexDump(Bytes(0x14))),
    )

class EndCmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x18, Int8ul),
        "unk_1" / Int8ul,
        "unk_2" / Int8ul,
        "flags" / Int8ul,
    )

    def __init__(self):
        super().__init__()
        self.unk_1 = 0
        self.unk_2 = 0
        self.flags = 0x40

class TimestampCmd(ConstructClass):
    subcon = Struct( # 0x34 bytes
        "magic" / Const(0x19, Int8ul),
        "unk_1" / Int8ul,
        "unk_2" / Int8ul,
        "unk_3" / Int8ul, # Sometimes 0x80
        # all these pointers point to 0xfa0... addresses. Might be where the timestamp should be writen?
        "ts0_addr" / Int64ul,
        "ts0" / ROPointer(this.ts0_addr, TimeStamp),
        "ts_pointers_addr" / Int64ul,
        "ts_pointers" / ROPointer(this.ts_pointers_addr, TimeStampPointers),
        "unk_addr" / Int64ul,
        "cmdqueue_ptr" / Int64ul,
        "user_ts_pointers_addr" / Int64ul,
        "ts_pointers" / ROPointer(this.user_ts_pointers_addr, TimeStampPointers),
        Ver("V >= V13_0B4", "unk_ts_addr" / Int64ul),
        "uuid" / Int32ul,
        "unk_30" / Int32ul,
    )

class WaitForInterruptCmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x01, Int8ul),
        "unk_1" / Int8ul,
        "unk_2" / Int8ul,
        "unk_3" / Int8ul,
    )

    def __init__(self, unk_1, unk_2, unk_3):
        super().__init__()
        self.unk_1 = unk_1
        self.unk_2 = unk_2
        self.unk_3 = unk_3

class Wait2Cmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x02, Int32ul),
    )

class NopCmd(ConstructClass):
    # This doesn't exist
    subcon = Struct(
        "magic" / Const(0x00, Int32ul),
    )

    def __str__(self) -> str:
        return "Nop"


class Write32Cmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x0e, Int8ul),
        "unk_1" / Int8ul,
        "unk_2" / Int8ul,
        "unk_3" / Int8ul,
        "addr" / Int64ul,
        "val" / Int32ul,
    )

    def __init__(self, addr, val):
        super().__init__()
        self.unk_1 = self.unk_2 = 0
        self.unk_3 = 0x20
        self.addr = addr
        self.val = val

class Store32Cmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x0e, Int8ul),
        "unk_1" / Int8ul,
        "unk_2" / Int8ul,
        "unk_3" / Int8ul,
        "addr" / Int64ul,
    )

    def __init__(self, addr):
        super().__init__()
        self.unk_1 = self.unk_2 = self.unk_3 = 0
        self.addr = addr

class Store64Cmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x0f, Int8ul),
        "unk_1" / Int8ul,
        "unk_2" / Int8ul,
        "unk_3" / Int8ul,
        "addr" / Int64ul,
    )

    def __init__(self, addr):
        super().__init__()
        self.unk_1 = self.unk_2 = self.unk_3 = 0
        self.addr = addr

class Write64Cmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x0f, Int8ul),
        "unk_1" / Int8ul,
        "unk_2" / Int8ul,
        "unk_3" / Int8ul,
        "addr" / Int64ul,
        "val" / Int64ul,
    )

    def __init__(self, addr, val):
        super().__init__()
        self.unk_1 = self.unk_2 = 0
        self.unk_3 = 0x20
        self.addr = addr
        self.val = val

class Read32Cmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x10, Int8ul),
        "unk_1" / Int8ul,
        "unk_2" / Int8ul,
        "unk_3" / Int8ul,
        "addr" / Int64ul,
    )

    def __init__(self, addr):
        super().__init__()
        self.unk_1 = self.unk_2 = self.unk_3 = 0
        self.addr = addr

class Read64Cmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x11, Int8ul),
        "unk_1" / Int8ul,
        "unk_2" / Int8ul,
        "unk_3" / Int8ul,
        "addr" / Int64ul,
    )

    def __init__(self, addr):
        super().__init__()
        self.unk_1 = self.unk_2 = self.unk_3 = 0
        self.addr = addr

class ALUCmd(ConstructClass):
    AND = 0
    OR = 1
    XOR = 2
    LSR = 3
    LSL = 4

    subcon = Struct(
        "magic" / Const(0x16, Int8ul),
        "v1" / Int8ul,
        "op" / Int16ul,
        "arg" / Int64ul,
    )

    def __init__(self, op, arg):
        super().__init__()
        self.v1 = 0
        self.op = (op << 3) | (2 << 11)
        self.arg = arg

class Add16Cmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x17, Int8ul),
        "arg1" / Int16ul,
        "arg2" / Int8ul,
    )

    def __init__(self, arg):
        super().__init__()
        self.arg1 = (arg << 3) & 0xffff
        self.arg2 = arg >> 13

class CompleteCmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x2b, Int8ul),
        ZPadding(3),
        "unk" / Int64ul,
        "stamp_addr" / Int64ul,
        "stamp_val" / Int32ul,
        "pad" / ZPadding(0x14),
    )

    def __init__(self):
        super().__init__()
        self.unk = 0

class AbortCmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x2c, Int8ul),
        ZPadding(3),
        "bits" / Int32ul,
        "unk1" / Int32ul,
        "unk2" / Int64ul,
        "unk3" / Int64ul,
        "stamp_addr" / Int64ul,
        "stamp_val" / Int32ul,
        "pad" / ZPadding(0x14),
    )

    def __init__(self):
        super().__init__()
        self.bits = 0
        self.unk1 = 0
        self.unk2 = 0
        self.unk3 = 0

class DoorbellCmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x03, Int8ul),
        "pad" / Int8ul,
        "arg" / Int16ul,
    )

    def __init__(self, flags):
        super().__init__()
        self.pad = 0
        self.arg = flags << 10

class ProfileCaptureCmd(ConstructClass):
    subcon = Struct(
        "magic" / Const(0x1f, Int8ul),
        "arg0" / Int8ul,
        "arg1" / Int8ul,
        "arg2" / Int8ul,
        "encoder_id" / Int32ul,
        "uuid" / Int32ul,
        "workitem_ptr" / Int64ul,
    )

    def __init__(self, flags):
        super().__init__()

class MicroSequence(ConstructValueClass):
    subcon = RepeatUntil(lambda obj, lst, ctx: lst[-1].op & 0x3f in (0x18, 0x2b, 0x2c),
                         Struct(
            "op" / Peek(Int32ul),
            #Probe(lookahead=32),
            "cmd" / Switch(this.op & 0x3f, {
                0x01: WaitForInterruptCmd,
                0x02: Wait2Cmd,
                0x03: DoorbellCmd,
                #0x04: write sgx u8
                #0x05: write sgx u32
                #0x06: write sgx u64
                #0x07: AGFSKUCommandTypeRegWriteStream not supported
                #0x08: read sgx u8
                #0x09: read sgx u32
                #0x0a: read sgx u64
                #0x0b: wait sgx reg u32
                #0x0c: wait sgx reg u64
                #0x0d: AGFSKUCommandTypeJump not supported
                0x0e:  Switch(this.op >> 28, {
                    0: Store32Cmd,
                    2: Write32Cmd,
                }),
                0x0f:  Switch(this.op >> 28, {
                    0: Store64Cmd,
                    2: Write64Cmd,
                }),
                0x10: Read32Cmd,
                0x11: Read64Cmd,
                0x16: ALUCmd,
                0x17: Add16Cmd,
                0x18: EndCmd,
                0x19: TimestampCmd,
                0x1f: ProfileCaptureCmd,
                #0x1a: KTraceCmd,
                0x22: StartTACmd,
                0x23: FinalizeTACmd,
                0x24: Start3DCmd,
                0x25: Finalize3DCmd,
                0x26: StartBlitCmd,
                0x27: FinalizeBlitCmd,
                0x29: StartComputeCmd,
                0x2a: FinalizeComputeCmd,
                0x2b: CompleteCmd,
                0x2c: AbortCmd,
            }, default=Error)
        )
    )

    def __str__(self):
        s = "{\n"
        for cmd in self.value:
            s += str(cmd.cmd) + '\n'
            if isinstance(cmd.cmd, (EndCmd, CompleteCmd)):
                s += "}\n"
                break
        else:
            s += "?\n"
        return s

__all__.extend(k for k, v in globals().items()
               if (callable(v) or isinstance(v, type)) and v.__module__ == __name__)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.constructutils import *
from m1n1.fw.agx import microsequence, initdata

#for v in initdata.__all__:
#for v in initdata.__dict__:
def dump(module):
    for v in module.__dict__:
        struct = getattr(module, v)
        if isinstance(struct, type) and issubclass(struct, ConstructClass) and struct is not ConstructClass:
            print(struct.to_rust())
            print()

dump(microsequence)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, fnmatch
import time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import struct
from m1n1.setup import *
from m1n1.fw.asc import StandardASC
from m1n1.hw.dart import DART
from m1n1.hw.dockchannel import DockChannel
from m1n1.fw.smc import SMCClient, SMCError
from m1n1.shell import run_shell
from m1n1.fw.mtp import *

from construct import *

smc_addr = u.adt["arm-io/smc"].get_reg(0)[0]
smc = SMCClient(u, smc_addr)
smc.start()
smc.start_ep(0x20)
smc.verbose = 0

p.dapf_init_all()

dart = DART.from_adt(u, "/arm-io/dart-mtp", iova_range=(0x8000, 0x100000))

dart.dart.regs.TCR[1].set(BYPASS_DAPF=0, BYPASS_DART=0, TRANSLATE_ENABLE=1)

irq_base = u.adt["/arm-io/dockchannel-mtp"].get_reg(1)[0]
fifo_base = u.adt["/arm-io/dockchannel-mtp"].get_reg(2)[0]
dc = DockChannel(u, irq_base, fifo_base, 1)

node = u.adt["/arm-io/dockchannel-mtp/mtp-transport"]

while dc.rx_count:
    dc.read(dc.rx_count)

mtp_addr = u.adt["/arm-io/mtp"].get_reg(0)[0]
mtp = StandardASC(u, mtp_addr, dart, stream=1)
mtp.boot()
mtp.verbose = 3
mtp.allow_phys = True
print("pre start")

def poll():
    mtp.work()
    mp.work_pending()

try:
    mp = MTPProtocol(u, node, mtp, dc, smc)

    mp.wait_init("keyboard")
    #mp.wait_init("multi_touch")
    mp.wait_init("stm")

    mtp.stop()
    mtp.start()
    mon.poll()

    for i in range(256):
        mp.stm.get_report(i)
        mtp.work()
        mp.work_pending()

    #for i in range(256):
        #if i in (0x40, 0x42):
            #continue
        #m = UnkDeviceControlMsg()
        #m.command = i
        #for args in (b"", b"\x00", b"\x01", b"\x02",
                     #b"\x01\x00", b"\x01\x01", b"\x01\x02",
                     #b"\x00\x01", b"\x00\x02", b"\x00\x00",
                     #b"\x00\x00\x00",
                     #b"\x00\x00\x00\x00",
                     #b"\x00\x00\x00\x00\x00",
                     #b"\x00\x00\x00\x00\x00\x00",
                     #b"\x00\x00\x00\x00\x00\x00\x00",
                     #b"\x00\x00\x00\x00\x00\x00\x00\x00",):
            #m.args = args
            #print(f"{m.command:#x} {m.args.hex()}")
            #mp.comm.device_control(m)

    #mon.poll()
    #mtp.stop()
    #mon.poll()
    #mtp.start()

    #mon.poll()
    #mtp.stop(1)
    ##reset(1)
    ##p.dapf_init_all()

    #mtp.boot()

    run_shell(locals(), poll_func=poll)

finally:
    #mtp.stop()
    p.reboot()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import argparse

parser = argparse.ArgumentParser(description='Enter a shell for codec-poking')
parser.add_argument('-n', '--no-reset', action="store_true")
args = parser.parse_args()

from m1n1.setup import *
from m1n1.shell import run_shell
from m1n1.hw.i2c import I2C, I2CRegMapDev
from m1n1.hw.codecs import *

i2c = {}

class TAS5770(I2CRegMapDev):
    REGMAP     = TAS5770Regs
    ADDRESSING = (1, 1)

class SN012776(I2CRegMapDev):
    REGMAP     = SN012776Regs
    ADDRESSING = (1, 1)

class CS42L84(I2CRegMapDev):
    REGMAP     = CS42L84Regs
    ADDRESSING = (0, 2)

class SSM3515(I2CRegMapDev):
    REGMAP     = SSM3515Regs
    ADDRESSING = (0, 1)

gpios = {}
for node in u.adt["/arm-io"]:
    if node.name.endswith("gpio") or node.name.endswith("gpio0"):
        gpios[node._properties["AAPL,phandle"]] = node

spks = []
for node in u.adt["/arm-io"]:
    if not node.name.startswith("i2c"):
        continue

    n = int(node.name[3:])
    i2c[n] = bus = I2C(u, f"/arm-io/{node.name}")

    for devnode in node:
        if "compatible" not in devnode._properties:
            continue

        dcls = {
            "audio-control,tas5770": TAS5770,
            "audio-control,sn012776": SN012776,
            "audio-control,cs42l84": CS42L84,
            "audio-control,ssm3515": SSM3515,
        }.get(devnode.compatible[0], None)

        if not dcls:
            continue

        dev = dcls.from_adt(bus, f"/arm-io/{node.name}/{devnode.name}")
        dev.node = devnode
        i2c[n].devs.append(dev)

        if type(dev) in [TAS5770, SN012776, SSM3515]:
            spks.append(dev)
        else:
            hp = dev

        if "function-reset" in devnode._properties:
            prop = devnode.function_reset
            gpio_host = gpios[prop.phandle]
            addr = gpio_host.get_reg(0)[0] + prop.args[0] * 4
            if not args.no_reset:
                print(f"Releasing #RST of {devnode.name}")
                p.mask32(addr, 1, 0)
            print(f"Pulling #RST of {devnode.name}")
            p.mask32(addr, 1, 1)

        if "interrupts" in devnode._properties \
                and devnode.interrupt_parent in gpios:
            gpio_host = gpios[devnode.interrupt_parent]
            addr = gpio_host.get_reg(0)[0] + devnode.interrupts[0] * 4
            print(f"Monitoring IRQ of {devnode.name}")
            mon.add(addr, 4)

run_shell(globals(), msg="Have fun!")



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, json, zipfile

json.c_make_encoder = None

from m1n1.proxy import *
from .context import *
from .event import GPUEventManager
from .uapi import *
from m1n1.constructutils import ConstructClass, Ver

def unswizzle(agx, addr, w, h, psize, dump=None, grid=False):
    iface = agx.u.iface

    tw = 64
    th = 64
    ntx = (w + tw - 1) // 64
    nty = (h + th - 1) // 64
    data = iface.readmem(addr, ntx * nty * psize * tw * th)
    new_data = []
    for y in range(h):
        ty = y // th
        for x in range(w):
            tx = x // tw
            toff = tw * th * psize * (ty * ntx + tx)
            j = x & (tw - 1)
            i = y & (th - 1)
            off = (
                ((j & 1) << 0) | ((i & 1) << 1) |
                ((j & 2) << 1) | ((i & 2) << 2) |
                ((j & 4) << 2) | ((i & 4) << 3) |
                ((j & 8) << 3) | ((i & 8) << 4) |
                ((j & 16) << 4) | ((i & 16) << 5) |
                ((j & 32) << 5) | ((i & 32) << 6))
            r,g,b,a = data[toff + psize*off: toff + psize*(off+1)]
            if grid:
                if x % 64 == 0 or y % 64 == 0:
                    r,g,b,a = 255,255,255,255
                elif x % 32 == 0 or y % 32 == 0:
                    r,g,b,a = 128,128,128,255
            new_data.append(bytes([b, g, r, a]))
    data = b"".join(new_data)
    if dump:
        open(dump, "wb").write(data[:w*h*psize])
    #iface.writemem(addr, data)

class GPUFrame:
    def __init__(self, context, filename=None, track=False):
        self.ctx = context
        self.agx = context.agx
        self.objects = []
        self.cmdbuf = None
        self.track = track
        if filename is not None:
            self.load(filename)

    def add_object(self, obj):
        self.objects.append(obj)

    def save(self, filename):
        cmdbuf = self.cmdbuf
        with zipfile.ZipFile(filename, "w") as zf:
            cmdbuf_data = json.dumps(cmdbuf, indent=4).encode("utf-8")
            zf.writestr("cmdbuf.json", cmdbuf_data)

            obj_info = []
            for obj in self.objects:
                if obj._data == bytes(obj._size):
                    filename = None
                else:
                    filename = f"obj_{obj._addr:x}.bin"
                    zf.writestr(filename, obj._data)
                obj_info.append({
                    "file": filename,
                    "name": obj._name,
                    "addr": obj._addr,
                    "size": obj._size,
                    "map_flags": obj._map_flags,
                })

            obj_info_data = json.dumps(obj_info, indent=4).encode("utf-8")
            zf.writestr("objects.json", obj_info_data)

    def load(self, filename):
        with zipfile.ZipFile(filename, "r") as zf:
            with zf.open("cmdbuf.json", "r") as fd:
                self.cmdbuf = drm_asahi_cmdbuf_t.from_json(fd)
            with zf.open("objects.json", "r") as fd:
                obj_info = json.load(fd)

            self.objects = []
            for i in obj_info:
                filename = i["file"]
                obj = self.ctx.new_at(i["addr"], Bytes(i["size"]), name=i["name"], track=self.track,
                                          **i["map_flags"])
                if filename is not None:
                    with zf.open(i["file"], "r") as fd:
                        data = fd.read()
                        obj.val = data
                        obj.push()
                else:
                    obj.val = bytes(i["size"])
                    obj.push()
                self.objects.append(obj)

class GPUWork:
    def __init__(self, renderer):
        self.objects = []
        self.renderer = renderer

    def add(self, obj):
        self.objects.append(obj)

    def free(self):
        for obj in self.objects:
            obj.free()
        self.objects = []

class GPURenderer:
    def __init__(self, ctx, buffers=16, bm_slot=0, queue=0):
        self.agx = agx = ctx.agx
        self.queue = queue
        self.scheduler_context = agx.kobj.new(GPUContextData).push()

        # 0..63
        self.ctx = ctx
        self.ctx_id = ctx.ctx

        # 0..255
        self.buffers = buffers
        self.buffer_mgr_slot = bm_slot

        ## These MUST go together
        self.buffer_mgr = GPUBufferManager(agx, ctx, buffers)
        self.buffer_mgr_initialized = False
        self.unk_emptybuf = agx.kobj.new_buf(0x40, "unk_emptybuf")
        self.tpc_size = 0

        ##### Job group

        self.job_list = agx.kshared.new(JobList)
        self.job_list.first_job = 0
        self.job_list.last_head = self.job_list._addr # Empty list has self as last_head
        self.job_list.unkptr_10 = 0
        self.job_list.push()

        ##### Work Queues

        self.wq_3d = GPU3DWorkQueue(agx, self.scheduler_context, self.job_list)
        self.wq_ta = GPUTAWorkQueue(agx, self.scheduler_context, self.job_list)

        self.wq_3d.info.uuid = 0x3D0000 | bm_slot
        self.wq_3d.info.push()
        self.wq_ta.info.uuid = 0x7A0000 | bm_slot
        self.wq_ta.info.push()

        self.stamp_value_3d = 0x3D000000 | (bm_slot << 16)
        self.stamp_value_ta = 0x7A000000 | (bm_slot << 16)

        ##### TA stamps

        # start?
        self.stamp_ta1 = agx.kshared.new(StampCounter, name="TA stamp 1")
        self.stamp_ta1.value = self.stamp_value_ta
        self.stamp_ta1.push()

        # complete?
        self.stamp_ta2 = agx.kobj.new(StampCounter, name="TA stamp 2")
        self.stamp_ta2.value = self.stamp_value_ta
        self.stamp_ta2.push()

        ##### 3D stamps

        # start?
        self.stamp_3d1 = agx.kshared.new(StampCounter, name="3D stamp 1")
        self.stamp_3d1.value = self.stamp_value_3d
        self.stamp_3d1.push()

        # complete?
        self.stamp_3d2 = agx.kobj.new(StampCounter, name="3D stamp 2")
        self.stamp_3d2.value = self.stamp_value_3d
        self.stamp_3d2.push()


        ##### Things userspace deals with for macOS

        #self.aux_fb = ctx.uobj.new_buf(0x8000, "Aux FB thing")
        ##self.deflake_1 = ctx.uobj.new_buf(0x20, "Deflake 1")
        ##self.deflake_2 = ctx.uobj.new_buf(0x280, "Deflake 2")
        ##self.deflake_3 = ctx.uobj.new_buf(0x540, "Deflake 3")
        #self.deflake = ctx.uobj.new_buf(0x7e0, "Deflake")
        #self.unk_buf = ctx.uobj.new(Array(0x800, Int64ul), "Unknown Buffer")
        #self.unk_buf.val = [0, *range(1, 0x400), *(0x400 * [0])]
        #self.unk_buf.push()

        ##### Some kind of feedback/status buffer, GPU managed?

        self.event_control = agx.kobj.new(EventControl)
        self.event_control.event_count = agx.kobj.new(Int32ul, "event_count")
        self.event_control.event_count.val = 0
        self.event_control.event_count.push()

        self.event_control.submission_id = 0
        self.event_control.cur_count = 0
        self.event_control.unk_10 = 0x50
        self.event_control.push()

        self.frames = 0

        self.ev_ta = ev_ta = self.agx.event_mgr.allocate_event()
        self.ev_3d = ev_3d = self.agx.event_mgr.allocate_event()

        self.work = []

        self.ev_idx = 0

        self.mshook_ta = None
        self.mshook_3d = None

    def submit(self, cmdbuf, wait_for=None):
        clustering = True
        nclusters = 8

        work = GPUWork(self)
        self.work.append(work)

        def mktimestamp(alloc, name):
            ts = alloc.new(Int64ul, name=name)
            ts.val = 0
            ts.push()
            work.add(ts)
            return ts

        work.ts3d_start = mktimestamp(self.agx.kshared, "3D timestamp start")
        work.ts3d_end = mktimestamp(self.agx.kshared, "3D timestamp end")
        work.tsta_start = mktimestamp(self.agx.kshared, "TA timestamp start")
        work.tsta_end = mktimestamp(self.agx.kshared, "TA timestamp end")

        work.ns3d_start = mktimestamp(self.agx.ktimestamp, "3D user timestamp start")
        work.ns3d_end = mktimestamp(self.agx.ktimestamp, "3D user timestamp end")
        work.nsta_start = mktimestamp(self.agx.ktimestamp, "TA timestamp start")
        work.nsta_end = mktimestamp(self.agx.ktimestamp, "TA timestamp end")

        self.buffer_mgr.increment()

        aux_fb = self.ctx.uobj.new_buf(0x20000, "Aux FB thing", track=False)
        work.add(aux_fb)

        # t8103
        deflake_1_size = 0x540
        deflake_2_size = 0x280
        deflake_3_size = 0x20

        # 4 * 0x40 on M2 Max?
        if Ver.check("G >= G14X"):
            deflake_3_size = 0x40

        # t6002 - 9 times larger instead of 8? works with 8...
        deflake_1_size *= nclusters
        deflake_2_size *= nclusters
        deflake_3_size *= nclusters

        deflake_1 = self.ctx.uobj.new_buf(deflake_1_size, "Deflake 1", track=False)
        deflake_2 = self.ctx.uobj.new_buf(deflake_2_size, "Deflake 2", track=False)
        deflake_3 = self.ctx.uobj.new_buf(deflake_3_size, "Deflake 3", track=False)
        work.add(deflake_1)
        work.add(deflake_2)
        work.add(deflake_3)

        unk_buf = self.ctx.uobj.new(Array(0x800, Int64ul), "Unknown Buffer", track=False)
        work.add(unk_buf)

        unk_buf.val = [0, *range(2, 0x401), *(0x400 * [0])]
        unk_buf.push()

        work.cmdbuf = cmdbuf

        self.frames += 1

        work.ev_ta = ev_ta = self.ev_ta
        work.ev_3d = ev_3d = self.ev_3d

        self.ev_ta.rearm()
        self.ev_3d.rearm()

        self.agx.log(f"ev_ta: {ev_ta.id}")
        self.agx.log(f"ev_3d: {ev_3d.id}")

        #self.event_control.base_stamp = self.stamp_value >> 8
        #self.event_control.push()

        self.prev_stamp_value_3d = self.stamp_value_3d
        self.prev_stamp_value_ta = self.stamp_value_ta
        self.stamp_value_3d += 0x100
        self.stamp_value_ta += 0x100
        self.event_control.event_count.val += 2
        self.event_control.event_count.push()

        work.stamp_value_3d = self.stamp_value_3d
        work.stamp_value_ta = self.stamp_value_ta

        agx = self.agx
        ctx = self.ctx

        work.width = width = cmdbuf.fb_width
        work.height = height = cmdbuf.fb_height

        ##### TVB allocations / Tiler config

        tile_width = 32
        tile_height = 32
        tiles_x = ((width + tile_width - 1) // tile_width)
        tiles_y = ((height + tile_height - 1) // tile_height)
        tiles = tiles_x * tiles_y

        mtiles_x = 4
        mtiles_y = 4

        mtile_x1 = align(((tiles_x + mtiles_x - 1) // mtiles_x), 4)
        mtile_x2 = 2 * mtile_x1
        mtile_x3 = 3 * mtile_x1
        mtile_y1 = align(((tiles_y + mtiles_y - 1) // mtiles_y), 4)
        mtile_y2 = 2 * mtile_y1
        mtile_y3 = 3 * mtile_y1

        mtile_stride = mtile_x1 * mtile_y1

        ## TODO: *samples
        tiles_per_mtile_x = mtile_x1
        tiles_per_mtile_y = mtile_y1

        tile_blocks_x = (tiles_x + 15) // 16
        tile_blocks_y = (tiles_y + 15) // 16
        tile_blocks = tile_blocks_x * tile_blocks_y

        tiling_params = TilingParameters()
        # rgn_header_size
        rgn_entry_size = 5
        tiling_params.size1 = (rgn_entry_size * tiles_per_mtile_x * tiles_per_mtile_y + 3) // 4
        # PPP_MULTISAMPLECTL
        tiling_params.unk_4 = 0x88
        # PPP_CTRL
        tiling_params.unk_8 = 0x203 # bit 0: GL clip mode
        # PPP_SCREEN
        tiling_params.x_max = width - 1
        tiling_params.y_max = height - 1
        # TE_SCREEN
        tiling_params.tile_count = ((tiles_y-1) << 12) | (tiles_x-1)
        # TE_MTILE1
        tiling_params.x_blocks = mtile_x3 | (mtile_x2 << 9) | (mtile_x1 << 18)
        # TE_MTILE2
        tiling_params.y_blocks = mtile_y3 | (mtile_y2 << 9) | (mtile_y1 << 18)
        tiling_params.size2 = mtile_stride
        tiling_params.size3 = 2 * mtile_stride
        tiling_params.unk_24 = 0x100
        tiling_params.unk_28 = 0x8000

        tilemap_size = (4 * tiling_params.size1 * mtiles_x * mtiles_y)

        tmtiles_x = tiles_per_mtile_x * mtiles_x
        tmtiles_y = tiles_per_mtile_y * mtiles_y

        tpc_entry_size = 8
        tpc_size = tpc_entry_size * tmtiles_x * tmtiles_y * nclusters

        if self.tpc_size < tpc_size:
            self.tpc = ctx.uobj.new_buf(tpc_size, "TPC", track=False).push()
            self.tpc_size = tpc_size

        depth_aux_buffer_addr = 0
        if cmdbuf.depth_buffer:
            size = align_pot(max(width, tile_width)) * align_pot(max(height, tile_width)) // 32
            depth_aux_buffer = self.ctx.uobj.new_buf(size, "Depth Aux", track=False)
            work.add(depth_aux_buffer)
            depth_aux_buffer_addr = depth_aux_buffer._addr

        stencil_aux_buffer_addr = 0
        if cmdbuf.stencil_buffer:
            size = align_pot(max(width, tile_width)) * align_pot(max(height, tile_width)) // 32
            stencil_aux_buffer = self.ctx.uobj.new_buf(size, "Stencil Aux", track=False)
            work.add(stencil_aux_buffer)
            stencil_aux_buffer_addr = stencil_aux_buffer._addr

        #tvb_tilemap_size = 0x80 * mtile_stride
        tvb_tilemap_size = tilemap_size
        tvb_tilemap = ctx.uobj.new_buf(tvb_tilemap_size, "TVB Tilemap", track=False).push()
        work.tvb_tilemap_size = tvb_tilemap_size
        work.tvb_tilemap = tvb_tilemap
        work.add(tvb_tilemap)

        # rogue: 0x180 * 4?
        tvb_heapmeta_size = 0x200
        #tvb_heapmeta_size = 0x600
        tvb_heapmeta = ctx.uobj.new_buf(tvb_heapmeta_size, "TVB Heap Meta", track=False).push()
        work.add(tvb_heapmeta)

        unk_tile_buf1 = self.ctx.uobj.new_buf(tvb_tilemap_size * nclusters, "Unk tile buf 1", track=False)
        print("tvb_tilemap_size", hex(tvb_tilemap_size))
        unk_tile_buf2 = self.ctx.uobj.new_buf(0x4 * nclusters, "Unk tile buf 2", track=False)
        #size = 0xc0 * nclusters
        size = 0xc80
        unk_tile_buf3 = self.ctx.uobj.new_buf(size, "Unk tile buf 3", track=False)
        f4 = 0x280
        #f4 = 0x400 # t602x?
        unk_tile_buf4 = self.ctx.uobj.new_buf(f4 * nclusters, "Unk tile buf 4", track=False)
        f5 = 0x30
        #f5 = 0x980
        unk_tile_buf5 = self.ctx.uobj.new_buf(f5 * nclusters, "Unk tile buf 5", track=False)
        work.add(unk_tile_buf1)
        work.add(unk_tile_buf2)
        work.add(unk_tile_buf3)
        work.add(unk_tile_buf4)
        work.add(unk_tile_buf5)

        ##### Buffer stuff?

        # buffer related?
        bufferthing_buf = ctx.uobj.new_buf(0x80, "BufferThing.unkptr_18", track=False)
        work.add(bufferthing_buf)

        work.buf_desc = buf_desc = agx.kobj.new(BufferThing, track=False)
        work.add(buf_desc)
        buf_desc.unk0_addr = self.buffer_mgr.get_scene()
        buf_desc.unk0_addr2 = buf_desc.unk0_addr
        buf_desc.unk_0 = 0x0
        buf_desc.unk_8 = 0x0
        buf_desc.unk_10 = 0x0
        buf_desc.unkptr_18 = bufferthing_buf._addr
        buf_desc.unk_20 = 0x0
        buf_desc.unk_28 = 0x0
        buf_desc.bm_misc_addr = self.buffer_mgr.misc_obj._addr
        buf_desc.unk_2c = 0x0
        buf_desc.unk_30 = 0x0
        buf_desc.unk_38 = 0x0
        buf_desc.push()

        uuid_3d = cmdbuf.cmd_3d_id
        uuid_ta = cmdbuf.cmd_ta_id
        encoder_id = cmdbuf.encoder_id

        #print(barrier_cmd)

        #self.wq_ta.submit(ta_barrier_cmd)

        ##### 3D barrier command

        barrier_cmd = agx.kobj.new(WorkCommandBarrier, track=False, align=0x20)
        work.add(barrier_cmd)
        barrier_cmd.stamp = self.stamp_ta2
        barrier_cmd.wait_value = self.stamp_value_ta
        barrier_cmd.stamp_self = self.stamp_value_3d
        barrier_cmd.event = ev_ta.id
        barrier_cmd.uuid = uuid_3d

        #print(barrier_cmd)

        self.wq_3d.submit(barrier_cmd)

        process_empty_tiles = True
        no_clear_pipeline_textures = True
        msaa_zs = False

        unk1 = False

        samples = 1
        layers = 1
        tile_config = 0x10000
        if not unk1:
            tile_config |= 0x280
        if layers > 1:
            tile_config |= 1
        if process_empty_tiles:
            tile_config |= 0x10000

        utile_config = 0xa000 | {1:0, 2:1, 4:2}[samples]
        ppp_multisamplectl = 0x88

        iogpu_unk_214 = 0xc000

        tib_blocks = 8
        large_tib = tib_blocks > 8
        set_when_reloading_z_or_s = False

        TAN_60 = 1.732051

        ##### 3D execution

        work.wc_3d = wc_3d = agx.cmdbuf.new(WorkCommand3D, track=False, align=0x20)
        work.add(work.wc_3d)
        wc_3d.counter = 0
        wc_3d.context_id = self.ctx_id
        wc_3d.unk_8 = 0
        wc_3d.event_control = self.event_control
        wc_3d.buffer_mgr = self.buffer_mgr.info
        wc_3d.buf_thing = buf_desc
        wc_3d.unk_emptybuf_addr = self.unk_emptybuf._addr
        wc_3d.tvb_tilemap = tvb_tilemap._addr
        wc_3d.unk_40 = ppp_multisamplectl
        wc_3d.unk_48 = samples
        wc_3d.tile_blocks_y = mtile_y1
        wc_3d.tile_blocks_x = mtile_x1
        wc_3d.unk_50 = 0x0
        wc_3d.unk_58 = 0x0

        wc_3d.merge_upper_x = TAN_60 / width
        wc_3d.merge_upper_y = TAN_60 / height
        wc_3d.unk_68 = 0x0
        wc_3d.tile_count = tiles

        wc_3d.unk_758 = Flag()
        wc_3d.unk_75c = Flag()
        wc_3d.unk_buf = WorkCommand1_UnkBuf()
        wc_3d.busy_flag = Flag()
        wc_3d.unk_buf2 = WorkCommand1_UnkBuf2()
        wc_3d.unk_buf2.unk_0 = 0
        wc_3d.unk_buf2.unk_8 = 0
        wc_3d.unk_buf2.unk_10 = 1
        wc_3d.ts1 = TimeStamp(0)
        wc_3d.ts_pointers = TimeStampPointers(work.ts3d_start._addr, work.ts3d_end._addr)
        wc_3d.user_ts_pointers = TimeStampPointers(work.ns3d_start._addr, work.ns3d_end._addr)
        wc_3d.client_sequence = 1
        # Ventura
        wc_3d.unk_ts = TimeStamp()
        wc_3d.unk_ts2 = TimeStamp()

        use_registers = Ver.check("G >= G14X")

        # Structures embedded in WorkCommand3D
        if not use_registers:
            wc_3d.struct_1 = Start3DStruct1()
            wc_3d.struct_1.store_pipeline_bind = cmdbuf.store_pipeline_bind
            wc_3d.struct_1.store_pipeline_addr = cmdbuf.store_pipeline | 4
            wc_3d.struct_1.unk_8 = 0x0
            wc_3d.struct_1.unk_c = 0x0

            wc_3d.struct_1.merge_upper_x = TAN_60 / width
            wc_3d.struct_1.merge_upper_y = TAN_60 / height

            wc_3d.struct_1.unk_18 = 0x0
            # ISP_MTILE_SIZE
            wc_3d.struct_1.tile_blocks_y = mtile_y1
            wc_3d.struct_1.tile_blocks_x = mtile_x1
            wc_3d.struct_1.unk_24 = 0x0
            wc_3d.struct_1.tile_counts = ((tiles_y-1) << 12) | (tiles_x-1)
            wc_3d.struct_1.unk_2c = tib_blocks
            wc_3d.struct_1.depth_clear_val1 = cmdbuf.depth_clear_value
            wc_3d.struct_1.stencil_clear_val1 = cmdbuf.stencil_clear_value
            wc_3d.struct_1.unk_35 = 0x7 # clear flags? 2 = depth 4 = stencil?
            wc_3d.struct_1.unk_36 = 0x0
            wc_3d.struct_1.unk_38 = 0x0
            wc_3d.struct_1.unk_3c = 0x1
            wc_3d.struct_1.unk_40 = 0
            wc_3d.struct_1.unk_44_padding = bytes(0x9c)

        if not use_registers:
            wc_3d.struct_2 = Start3DStruct2()
            wc_3d.struct_2.unk_0 = utile_config
            wc_3d.struct_2.clear_pipeline = Start3DClearPipelineBinding(
                cmdbuf.load_pipeline_bind, cmdbuf.load_pipeline | 4)
            wc_3d.struct_2.unk_18 = ppp_multisamplectl
            wc_3d.struct_2.scissor_array = cmdbuf.scissor_array
            wc_3d.struct_2.depth_bias_array = cmdbuf.depth_bias_array
            wc_3d.struct_2.aux_fb = AuxFBInfo(iogpu_unk_214, 0, width, height)
            # ISP_ZLS_PIXELS
            wc_3d.struct_2.depth_dimensions = (width - 1) | ((height - 1) << 15)
            wc_3d.struct_2.visibility_result_buffer = 0x0
            # ISP_ZLSCTL
            wc_3d.struct_2.depth_flags = cmdbuf.ds_flags
            wc_3d.struct_2.unk_58_g14_0 = 0x4040404
            wc_3d.struct_2.unk_58_g14_8 = 0
            wc_3d.struct_2.depth_buffer_ptr1 = cmdbuf.depth_buffer
            wc_3d.struct_2.depth_buffer_ptr2 = cmdbuf.depth_buffer
            wc_3d.struct_2.unk_68_g14_0 = 0
            wc_3d.struct_2.stencil_buffer_ptr1 = cmdbuf.stencil_buffer
            wc_3d.struct_2.stencil_buffer_ptr2 = cmdbuf.stencil_buffer
            wc_3d.struct_2.unk_78 = [0] * 4
            wc_3d.struct_2.depth_aux_buffer_ptr1 = depth_aux_buffer_addr
            wc_3d.struct_2.unk_a0 = 0
            wc_3d.struct_2.depth_aux_buffer_ptr2 = depth_aux_buffer_addr
            wc_3d.struct_2.unk_b0 = 0
            wc_3d.struct_2.stencil_aux_buffer_ptr1 = stencil_aux_buffer_addr
            wc_3d.struct_2.unk_c0 = 0
            wc_3d.struct_2.stencil_aux_buffer_ptr2 = stencil_aux_buffer_addr
            wc_3d.struct_2.unk_d0 = 0
            wc_3d.struct_2.tvb_tilemap = tvb_tilemap._addr
            wc_3d.struct_2.tvb_heapmeta_addr = tvb_heapmeta._addr
            wc_3d.struct_2.unk_e8 = tiling_params.size1 << 24
            wc_3d.struct_2.tvb_heapmeta_addr2 = tvb_heapmeta._addr
            # 0x10000 - clear empty tiles
            # ISP_CTL (but bits seem to have moved)
            wc_3d.struct_2.unk_f8 = tile_config
            wc_3d.struct_2.aux_fb_ptr = aux_fb._addr
            wc_3d.struct_2.unk_108 = [0x0, 0x0, 0x0, 0x0, 0x0, 0x0]
            wc_3d.struct_2.pipeline_base = self.ctx.pipeline_base
            wc_3d.struct_2.unk_140 = 0x8c60
            wc_3d.struct_2.helper_program = 0
            wc_3d.struct_2.unk_14c = 0
            wc_3d.struct_2.helper_arg = 0
            wc_3d.struct_2.unk_158 = 0x1c
            wc_3d.struct_2.unk_160 = 0
            wc_3d.struct_2.unk_168_padding = bytes(0x1d8)
            wc_3d.struct_2.unk_198_padding = bytes(0x1a8)

        def fui(v):
            return struct.unpack("<I", struct.pack("<f", v))[0]

        if use_registers:
            regs = {
                0x1739:     1,
                0x10009:    utile_config,                       # s2.unk_0 utile_config
                0x15379:    cmdbuf.store_pipeline_bind,         # s1.store_pipeline_bind
                0x15381:    cmdbuf.store_pipeline | 4,          # s1.store_pipeline_addr
                0x15369:    cmdbuf.load_pipeline_bind,          # s2.clear_pipeline.bind
                0x15371:    cmdbuf.load_pipeline | 4,           # s2.clear_pipeline.addr
                0x15131:    fui(TAN_60 / width),                # ISP_MERGE_UPPER_X
                0x15139:    fui(TAN_60 / height),               # ISP_MERGE_UPPER_Y
                0x100a1:    0x0,
                0x15069:    0,
                0x15071:    0,                                  # pointer
                0x16058:    0,
                0x10019:    ppp_multisamplectl,                 # PPP_MULTISAMPLECTL
                0x100b1:    mtile_y1 | (mtile_x1 << 16),        # ISP_MTILE_SIZE
                0x16030:    mtile_y1 | (mtile_x1 << 16),        # ISP_MTILE_SIZE
                0x100d9:    ((tiles_y-1) << 12) | (tiles_x-1),  # TE_SCREEN
                0x16098:    tvb_heapmeta._addr,                 # tvb_heapmeta_addr?
                0x15109:    cmdbuf.scissor_array,               # ISP_SCISSOR_BASE
                0x15101:    cmdbuf.depth_bias_array,            # ISP_DBIAS_BASE
                0x15021:    0,                                  # s2.aux_fb.unk_1
                0x15211:    (height << 32) | width,             # s2.aux_fb.{width, height}
                0x15049:    0,                                  # s2.aux_fb.unk3
                0x10051:    tib_blocks,                         # s1.unk_2c
                0x15321:    (width - 1) | ((height - 1) << 15), # ISP_ZLS_PIXELS
                0x15301:    fui(cmdbuf.depth_clear_value),      # ISP_BGOBJDEPTH
                0x15309:    cmdbuf.stencil_clear_value | 0x300, # ISP_BGOBJVALS
                0x15311:    0,                                  # ISP_OCLQRY_BASE
                0x15319:    cmdbuf.ds_flags,                    # ISP_ZLSCTL
                0x15349:    0x4040404,                          # s2.unk_58_g14_0
                0x15351:    0,                                  # s2.unk_58_g14_8
                0x15329:    cmdbuf.depth_buffer,                # ISP_ZLOAD_BASE
                0x15331:    cmdbuf.depth_buffer,                # ISP_ZSTORE_BASE
                0x15339:    cmdbuf.stencil_buffer,              # ISP_STENCIL_LOAD_BASE
                0x15341:    cmdbuf.stencil_buffer,              # ISP_STENCIL_STORE_BASE
                0x15231:    0,
                0x15221:    0,
                0x15239:    0,
                0x15229:    0,
                0x15401:    0,
                0x15421:    0,
                0x15409:    0,
                0x15429:    0,
                0x153c1:    depth_aux_buffer_addr,
                0x15411:    0,
                0x153c9:    depth_aux_buffer_addr,
                0x15431:    0,
                0x153d1:    stencil_aux_buffer_addr,
                0x15419:    0,
                0x153d9:    stencil_aux_buffer_addr,
                0x15439:    0,
                0x16429:    tvb_tilemap._addr,                  # tvb_tilemap
                0x16060:    tvb_heapmeta._addr,                 # tvb_heapmeta_addr
                0x16431:    (4 * tiling_params.size1) << 24,    # s2.unk_e8
                0x10039:    tile_config,                        # tile_config ISP_CTL?
                0x16451:    0x0,                                # tile offset Y/X
                0x11821:    0x0,                                # some shader?
                0x11829:    0,
                0x11f79:    0,
                0x15359:    0,
                0x10069:    self.ctx.pipeline_base,             # USC_EXEC_BASE_ISP
                0x16020:    0,
                0x16461:    aux_fb._addr,                       # aux_fb_ptr?
                0x16090:    aux_fb._addr,                       # aux_fb_ptr?
                0x120a1:    0x1c,                               # s2.unk_158
                0x160a8:    0,
                0x16068:    (
                                0x0000000_00036011 |
                                ((tiles_x-1) << 44) |
                                ((tiles_y-1) << 53) |
                                (0 if unk1 else 0x20_00000000) |
                                ((utile_config & 0xf000) << 28)
                            ),
                0x160b8:    0x0,
            }

            if False:
                # These are actually appended by the firmware?
                regs[0x10211] = 0x134
                regs[0x10420] = 0x134
                regs[0x1c838] = 0x1
                regs[0x1ca28] = 0x1502961540 # buf_desc.unkptr_18
                regs[0x1731] = 0x1

            wc_3d.registers = [RegisterDefinition(k, v) for k, v in regs.items()]
            reg_count = len(wc_3d.registers)
            wc_3d.registers += [RegisterDefinition(0, 0) for i in range(len(regs), 128)]

        wc_3d.set_addr() # Update inner structure addresses

        if True:
            wc_3d.struct_3 = Start3DStruct3()
            if use_registers:
                wc_3d.struct_3.registers_addr = wc_3d.registers[0]._addr
                wc_3d.struct_3.register_count = reg_count
                wc_3d.struct_3.registers_length = reg_count * 12
            else:
                wc_3d.struct_3.registers_addr = 0
                wc_3d.struct_3.register_count = 0
                wc_3d.struct_3.registers_length = 0
            wc_3d.struct_3.unk_d8 = 0
            wc_3d.struct_3.depth_bias_array = Start3DArrayAddr(cmdbuf.depth_bias_array)
            wc_3d.struct_3.scissor_array = Start3DArrayAddr(cmdbuf.scissor_array)
            wc_3d.struct_3.visibility_result_buffer = 0x0
            wc_3d.struct_3.unk_118 = 0x0
            wc_3d.struct_3.unk_120 = [0] * 37
            wc_3d.struct_3.unk_reload_pipeline = Start3DClearPipelineBinding(
                cmdbuf.partial_reload_pipeline_bind, cmdbuf.partial_reload_pipeline | 4)
            wc_3d.struct_3.unk_258 = 0
            wc_3d.struct_3.unk_260 = 0
            wc_3d.struct_3.unk_268 = 0
            wc_3d.struct_3.unk_270 = 0
            wc_3d.struct_3.reload_pipeline = Start3DClearPipelineBinding(
                cmdbuf.partial_reload_pipeline_bind, cmdbuf.partial_reload_pipeline | 4)
            wc_3d.struct_3.depth_flags = cmdbuf.ds_flags | 0x44
            wc_3d.struct_3.unk_290 = 0x4040404
            wc_3d.struct_3.depth_buffer_ptr1 = cmdbuf.depth_buffer
            wc_3d.struct_3.unk_2a0 = 0x0
            wc_3d.struct_3.unk_2a8 = 0x0
            wc_3d.struct_3.depth_buffer_ptr2 = cmdbuf.depth_buffer
            wc_3d.struct_3.depth_buffer_ptr3 = cmdbuf.depth_buffer
            wc_3d.struct_3.depth_aux_buffer_ptr = depth_aux_buffer_addr
            wc_3d.struct_3.stencil_buffer_ptr1 = cmdbuf.stencil_buffer
            wc_3d.struct_3.unk_2d0 = 0x0
            wc_3d.struct_3.unk_2d8 = 0x0
            wc_3d.struct_3.stencil_buffer_ptr2 = cmdbuf.stencil_buffer
            wc_3d.struct_3.stencil_buffer_ptr3 = cmdbuf.stencil_buffer
            wc_3d.struct_3.stencil_aux_buffer_ptr = stencil_aux_buffer_addr
            wc_3d.struct_3.unk_2f8 = [0x0, 0x0]
            wc_3d.struct_3.aux_fb_unk0 = tib_blocks
            wc_3d.struct_3.unk_30c = 0x0
            wc_3d.struct_3.aux_fb = AuxFBInfo(iogpu_unk_214, 0, width, height)
            if use_registers:
                wc_3d.struct_3.s2_unk_f8 = tile_config
            else:
                wc_3d.struct_3.s2_unk_f8 = 0
            wc_3d.struct_3.unk_324_padding = bytes(0xc)
            wc_3d.struct_3.unk_partial_store_pipeline = Start3DStorePipelineBinding(
                cmdbuf.partial_store_pipeline_bind, cmdbuf.partial_store_pipeline | 4)
            wc_3d.struct_3.partial_store_pipeline = Start3DStorePipelineBinding(
                cmdbuf.partial_store_pipeline_bind, cmdbuf.partial_store_pipeline | 4)
            wc_3d.struct_3.depth_clear_val2 = cmdbuf.depth_clear_value
            wc_3d.struct_3.stencil_clear_val2 = cmdbuf.stencil_clear_value
            wc_3d.struct_3.unk_375 = 3
            wc_3d.struct_3.unk_376 = 0x0
            wc_3d.struct_3.unk_378 = 0x10
            wc_3d.struct_3.unk_37c = 0x0
            wc_3d.struct_3.unk_380 = 0x0
            wc_3d.struct_3.unk_388 = 0x0
            wc_3d.struct_3.unk_390_0 = 0x0 # Ventura
            wc_3d.struct_3.depth_dimensions = (width - 1) | ((height - 1) << 15)

        if True:
            wc_3d.struct_6 = Start3DStruct6()
            wc_3d.struct_6.tvb_overflow_count = 0x0
            wc_3d.struct_6.unk_8 = int(set_when_reloading_z_or_s)
            wc_3d.struct_6.unk_c = int(large_tib)
            wc_3d.struct_6.unk_10 = 0x0
            wc_3d.struct_6.encoder_id = cmdbuf.encoder_id
            wc_3d.struct_6.unk_1c = 0xffffffff
            wc_3d.struct_6.sampler_array = unk_buf._addr
            wc_3d.struct_6.sampler_count = 0x0
            wc_3d.struct_6.sampler_max = 0x0
            wc_3d.struct_6.unk_30 = int(process_empty_tiles)
            wc_3d.struct_6.unk_34 = int(no_clear_pipeline_textures)
            wc_3d.struct_6.unk_38 = int(msaa_zs)

        if True:
            wc_3d.struct_7 = Start3DStruct7()
            wc_3d.struct_7.unk_0_0 = 0x0
            wc_3d.struct_7.unk_0 = 0x0
            wc_3d.struct_7.stamp1 = self.stamp_3d1
            wc_3d.struct_7.stamp2 = self.stamp_3d2
            wc_3d.struct_7.stamp_value = self.stamp_value_3d
            wc_3d.struct_7.ev_3d = ev_3d.id
            wc_3d.struct_7.evctl_index = self.ev_idx
            wc_3d.struct_7.unk_24 = 1
            wc_3d.struct_7.uuid = uuid_3d
            wc_3d.struct_7.queue_cmd_count = 0
            wc_3d.struct_7.unk_30 = int(unk1)

        wc_3d.set_addr() # Update inner structure addresses
        #print("WC3D", hex(wc_3d._addr))
        #print(" s1", hex(wc_3d.struct_1._addr))
        #print(" s2", hex(wc_3d.struct_2._addr))
        #print(" s6", hex(wc_3d.struct_6._addr))
        #print(" s7", hex(wc_3d.struct_7._addr))

        ms = GPUMicroSequence(agx)
        if self.mshook_3d:
            self.mshook_3d(self, work, ms)

        start_3d = Start3DCmd()
        if not use_registers:
            start_3d.struct1_addr = wc_3d.struct_1._addr # 0x44 bytes!
            start_3d.struct2_addr = wc_3d.struct_2._addr # 0x168 bytes!
            start_3d.registers_addr = 0
        else:
            start_3d.struct1_addr = 0
            start_3d.struct2_addr = 0
            start_3d.registers_addr = wc_3d.registers[0]._addr
        start_3d.buf_thing = buf_desc
        start_3d.stats_ptr = agx.initdata.regionB.stats_3d.stats._addr
        start_3d.busy_flag_ptr = wc_3d.busy_flag._addr
        start_3d.struct6 = wc_3d.struct_6 # 4 bytes!
        start_3d.struct7 = wc_3d.struct_7 # 4 bytes!
        start_3d.cmdqueue_ptr = self.wq_3d.info._addr
        start_3d.workitem_ptr = wc_3d._addr
        start_3d.context_id = self.ctx_id
        start_3d.unk_50 = 0x1
        start_3d.submission_id = self.event_control.submission_id
        start_3d.buffer_mgr_slot = self.buffer_mgr_slot
        start_3d.unk_5c = int(large_tib)
        start_3d.queue_cmd_count = self.prev_stamp_value_3d >> 8
        start_3d.unk_68 = 0x0
        start_3d.unk_buf_ptr = wc_3d.unk_758._addr
        start_3d.unk_buf2_ptr = wc_3d.unk_buf2._addr
        start_3d.unk_7c_0 = 0x0
        start_3d.unk_7c = 0x0
        start_3d.unk_80 = 0x0
        start_3d.unk_84 = int(unk1)
        start_3d.uuid = uuid_3d
        start_3d.attachments = []
        start_3d.counter = wc_3d.counter
        start_3d.unkptr_19c = self.event_control.unk_buf._addr

        work.fb = None
        work.depth = None

        for i in cmdbuf.attachments[:cmdbuf.attachment_count]:
            cache_lines = align_up(i.size, 128) // 128
            order = 1 # FIXME
            start_3d.attachments.append(Attachment(i.pointer, cache_lines, 0x17, order)) # FIXME check
            if work.fb is None and i.type == ASAHI_ATTACHMENT_C:
                work.fb = i.pointer
            if work.depth is None and i.type == ASAHI_ATTACHMENT_Z:
                work.depth = i.pointer
        start_3d.attachments += [Attachment(0, 0, 0, 0)] * (16 - len(start_3d.attachments))
        start_3d.num_attachments = cmdbuf.attachment_count
        start_3d.unk_190 = 0x0

        start_3d_offset = ms.append(start_3d)

        ts1 = TimestampCmd()
        ts1.unk_1 = 0x0
        ts1.unk_2 = 0x0
        ts1.unk_3 = 0x80
        ts1.ts0_addr = wc_3d.ts1._addr
        ts1.ts_pointers_addr = wc_3d.ts_pointers._addr
        ts1.unk_addr = 0xdeadbeef
        ts1.cmdqueue_ptr = self.wq_3d.info._addr
        ts1.user_ts_pointers_addr = wc_3d.user_ts_pointers._addr
        if Ver.check("V >= V13_0B4"):
            ts1.unk_ts_addr = wc_3d.unk_ts._addr
        ts1.uuid = uuid_3d
        ts1.unk_30 = 0x0
        ms.append(ts1)

        if Ver.check("G >= G14X"):
            ms.append(Wait2Cmd())
        else:
            ms.append(WaitForInterruptCmd(0, 1, 0))

        ts2 = TimestampCmd()
        ts2.unk_1 = 0x0
        ts2.unk_2 = 0x0
        ts2.unk_3 = 0x0
        ts2.ts0_addr = wc_3d.ts1._addr
        ts2.ts_pointers_addr = wc_3d.ts_pointers._addr
        ts2.unk_addr = 0xdeadbeef
        ts2.cmdqueue_ptr = self.wq_3d.info._addr
        ts2.user_ts_pointers_addr = wc_3d.user_ts_pointers._addr
        if Ver.check("V >= V13_0B4"):
            ts2.unk_ts_addr = wc_3d.unk_ts._addr
        ts2.uuid = uuid_3d
        ts2.unk_30 = 0x0
        ms.append(ts2)

        finish_3d = Finalize3DCmd()
        finish_3d.uuid = uuid_3d
        finish_3d.unk_8 = 0
        finish_3d.stamp = self.stamp_3d2
        finish_3d.stamp_value = self.stamp_value_3d
        finish_3d.unk_18 = 0
        finish_3d.buf_thing = buf_desc
        finish_3d.buffer_mgr = self.buffer_mgr.info
        finish_3d.unk_2c = 1
        finish_3d.stats_ptr = agx.initdata.regionB.stats_3d.stats._addr
        finish_3d.struct7 = wc_3d.struct_7
        finish_3d.busy_flag_ptr = wc_3d.busy_flag._addr
        finish_3d.cmdqueue_ptr = self.wq_3d.info._addr
        finish_3d.workitem_ptr = wc_3d._addr
        finish_3d.unk_5c = self.ctx_id
        finish_3d.unk_buf_ptr = wc_3d.unk_758._addr
        finish_3d.unk_6c_0 = 0
        finish_3d.unk_6c = 0
        finish_3d.unk_74 = 0
        finish_3d.unk_7c = 0
        finish_3d.unk_84 = 0
        finish_3d.unk_8c = 0
        finish_3d.unk_8c_g14 = 0
        finish_3d.restart_branch_offset = start_3d_offset - ms.off
        finish_3d.unk_98 = 0
        finish_3d.unk_9c = bytes(0x10)
        ms.append(finish_3d)
        ms.finalize()

        work.add(ms.obj)

        wc_3d.microsequence_ptr = ms.obj._addr
        wc_3d.microsequence_size = ms.size

        print(wc_3d)
        self.wq_3d.submit(wc_3d)

        ##### TA init

        #print(ctx_info)
        if wait_for is not None:
            barrier_cmd = agx.kobj.new(WorkCommandBarrier, track=False, align=0x20)
            work.add(barrier_cmd)
            if not isinstance(wait_for, tuple):
                barrier_cmd.stamp = wait_for.renderer.stamp_3d2
                barrier_cmd.wait_value = wait_for.stamp_value_3d
                barrier_cmd.event = wait_for.ev_3d.id
            else:
                barrier_cmd.stamp_addr = wait_for[0]
                barrier_cmd.wait_value = wait_for[1]
                barrier_cmd.event = wait_for[2]

            barrier_cmd.stamp_self = self.stamp_value_ta
            barrier_cmd.uuid = uuid_ta

            self.wq_ta.submit(barrier_cmd)

        if not self.buffer_mgr_initialized:
            wc_initbm = agx.kobj.new(WorkCommandInitBM, track=False, align=0x20)
            work.add(wc_initbm)
            wc_initbm.context_id = self.ctx_id
            wc_initbm.buffer_mgr_slot = self.buffer_mgr_slot
            wc_initbm.unk_c = 0
            wc_initbm.unk_10 = self.buffer_mgr.info.block_count
            wc_initbm.buffer_mgr = self.buffer_mgr.info
            wc_initbm.stamp_value = self.stamp_value_ta

            self.wq_ta.submit(wc_initbm)

            self.buffer_mgr_initialized = True

        ##### TA execution

        work.wc_ta = wc_ta = agx.cmdbuf.new(WorkCommandTA, track=False, align=0x20)
        work.add(work.wc_ta)
        wc_ta.context_id = self.ctx_id
        wc_ta.counter = 1
        wc_ta.unk_8 = 0
        wc_ta.event_control = self.event_control
        wc_ta.buffer_mgr_slot = self.buffer_mgr_slot
        wc_ta.buffer_mgr = self.buffer_mgr.info
        wc_ta.buf_thing = buf_desc
        wc_ta.unk_emptybuf_addr = wc_3d.unk_emptybuf_addr
        wc_ta.unk_34 = 0x0

        wc_ta.unk_3e8 = bytes(0x64)
        wc_ta.unk_594 = WorkCommand0_UnkBuf()

        wc_ta.ts1 = TimeStamp(0)
        wc_ta.ts_pointers = TimeStampPointers(work.tsta_start._addr, work.tsta_end._addr)
        wc_ta.user_ts_pointers = TimeStampPointers(work.nsta_start._addr, work.nsta_end._addr)
        wc_ta.client_sequence = 1
        # Ventura
        wc_ta.unk_ts = TimeStamp()
        wc_ta.unk_ts2 = TimeStamp()

        # Structures embedded in WorkCommandTA
        if not use_registers:
            wc_ta.tiling_params = tiling_params

        import random

        tiling_control = agx.chip_info.tiling_control
        if not clustering:
            tiling_control |= 1

        if not use_registers:
            wc_ta.struct_2 = StartTACmdStruct2()
            wc_ta.struct_2.unk_0 = 0 if unk1 else 0x200
            wc_ta.struct_2.unk_8 = 0x1e3ce508 # fixed
            wc_ta.struct_2.unk_c = 0x1e3ce508 # fixed
            wc_ta.struct_2.tvb_tilemap = tvb_tilemap._addr
            wc_ta.struct_2.tvb_cluster_tilemaps = unk_tile_buf1._addr
            wc_ta.struct_2.tpc = self.tpc._addr
            wc_ta.struct_2.tvb_heapmeta_addr = tvb_heapmeta._addr | 0x8000_0000_0000_0000
            wc_ta.struct_2.iogpu_unk_54 = 0x6b0003 # fixed
            wc_ta.struct_2.iogpu_unk_55 = 0x3a0012 # fixed
            wc_ta.struct_2.iogpu_unk_56 = 0x1 # fixed
            wc_ta.struct_2.tvb_cluster_meta1 = unk_tile_buf2._addr | 0x4_0000_0000_0000
            wc_ta.struct_2.unk_48 = utile_config
            wc_ta.struct_2.unk_50 = 0x88 # fixed
            wc_ta.struct_2.tvb_heapmeta_addr2 = tvb_heapmeta._addr
            wc_ta.struct_2.unk_60 = 0x0 # fixed
            wc_ta.struct_2.core_mask = 0xffffffffffffffff
            #wc_ta.struct_2.unk_68 = 0xff << (8 * (self.buffer_mgr_slot % 8))
            wc_ta.struct_2.iogpu_deflake_1 = deflake_1._addr
            wc_ta.struct_2.iogpu_deflake_2 = deflake_2._addr
            wc_ta.struct_2.unk_80 = 0x1 # fixed
            wc_ta.struct_2.iogpu_deflake_3 = deflake_3._addr | 0x4_0000_0000_0000 # check
            wc_ta.struct_2.encoder_addr = cmdbuf.encoder_ptr
            wc_ta.struct_2.tvb_cluster_meta2 = unk_tile_buf3._addr
            wc_ta.struct_2.tvb_cluster_meta3 = unk_tile_buf4._addr
            wc_ta.struct_2.tiling_control = tiling_control
            wc_ta.struct_2.unk_b0 = [0x0, 0x0, 0x0, 0x0, 0x0, 0x0] # fixed
            wc_ta.struct_2.pipeline_base = self.ctx.pipeline_base
            wc_ta.struct_2.tvb_cluster_meta4 = unk_tile_buf5._addr | 0x3000_0000_0000_0000
            wc_ta.struct_2.unk_f0 = 0x20 # fixed
            wc_ta.struct_2.unk_f8 = 0x8c60 # fixed
            wc_ta.struct_2.helper_program = 0
            wc_ta.struct_2.unk_104 = 0
            wc_ta.struct_2.helper_arg = 0
            wc_ta.struct_2.unk_110 = 0
            wc_ta.struct_2.unk_118 = 0x1c # fixed

            wc_ta.registers_addr = 0
            wc_ta.register_count = 0
            wc_ta.registers_length = 0
        else:
            regs = {
                0x10141:    (0 if unk1 else 0x200),               # s2.unk_0
                0x1c039:    tvb_tilemap._addr,
                0x1c9c8:    tvb_tilemap._addr,
                0x1c041:    unk_tile_buf1._addr,                # s2.tvb_cluster_tilemaps
                0x1c9d0:    unk_tile_buf1._addr,
                0x1c0a1:    self.tpc._addr,                     # TE_TPC_ADDR
                0x1c031:    tvb_heapmeta._addr | 0x8000_0000_0000_0000,
                0x1c9c0:    tvb_heapmeta._addr | 0x8000_0000_0000_0000,
                0x1c051:    0x3a0012006b0003,                   # iogpu_unk_54/55
                0x1c061:    1,                                  # iogpu_unk_56
                0x10149:    utile_config,                       # s2.unk_48 utile_config
                0x10139:    ppp_multisamplectl,                 # PPP_MULTISAMPLECTL
                0x10111:    deflake_1._addr,
                0x1c9b0:    deflake_1._addr,
                0x10119:    deflake_2._addr,
                0x1c9b8:    deflake_2._addr,
                0x1c958:    1,                                  # s2.unk_80
                0x1c950:    deflake_3._addr | 0x4_0000_0000_0000,
                0x1c930:    0,                                  # VCE related addr, lsb to enable
                0x1c880:    cmdbuf.encoder_ptr,                 # VDM_CTRL_STREAM_BASE
                0x1c898:    0x0,                                # if lsb set, faults in UL1C0, possibly missing addr.
                0x1c948:    unk_tile_buf3._addr,                # tvb_cluster_meta2
                0x1c888:    unk_tile_buf4._addr,                # tvb_cluster_meta3
                0x1c890:    tiling_control,                     # tvb_tiling_control
                0x1c918:    0x4,
                0x1c079:    tvb_heapmeta._addr,
                0x1c9d8:    tvb_heapmeta._addr,
                0x1c089:    0,
                0x1c9e0:    0,
                0x16c41:    unk_tile_buf5._addr,                # tvb_cluster_meta4
                0x1ca40:    unk_tile_buf5._addr,                # tvb_cluster_meta4
                0x1c9a8:    0x1c, # + meta1_blocks?             # min_free_tvb_pages?
                0x1c920:    unk_tile_buf2._addr,                # ??? | meta1_blocks?
                0x10151:    0,
                0x1c199:    0,
                0x1c1a1:    0,
                0x1c1a9:    0,                                  # 0x10151 bit 1 enables
                0x1c1b1:    0,
                0x1c1b9:    0,
                0x10061:    self.ctx.pipeline_base,             # USC_EXEC_BASE_TA
                0x11801:    0,                                  # some shader?
                0x11809:    0,                                  # maybe arg?
                0x11f71:    0,
                0x1c0b1:    tiling_params.size1,                # TE_PSG
                0x1c850:    tiling_params.size1,
                0x10131:    tiling_params.unk_4,
                0x10121:    tiling_params.unk_8,                # PPP_CTRL
                0x10129:    tiling_params.x_max | (tiling_params.y_max << 16),  # PPP_SCREEN
                0x101b9:    tiling_params.tile_count,           # TE_SCREEN
                0x1c069:    tiling_params.x_blocks,             # TE_MTILE1
                0x1c071:    tiling_params.y_blocks,             # TE_MTILE2
                0x1c081:    tiling_params.size2,                # TE_MTILE
                0x1c0a9:    tiling_params.size3,                # TE_TPC
                0x10171:    tiling_params.unk_24,
                0x10169:    tiling_params.unk_28,               # TA_RENDER_TARGET_MAX
                0x12099:    0,
                0x1c9e8:    0,
            }

            if False:
                # These are actually appended by the firmware?
                regs[0x10209] = 0x133
                regs[0x1c9f0] = 0x133
                regs[0x1c830] = 0x1
                regs[0x1c9f0] = 0x1502960fa0
                regs[0x16c39] = 0x1502960fa0 # buf_desc.unkptr_18
                regs[0x1c910] = 0xa0000b0125
                regs[0x1c8e0] = 0xff # core_mask_0
                regs[0x1c8e8] = 0x00 # core_mask_1

            wc_ta.registers = [RegisterDefinition(k, v) for k, v in regs.items()]
            ta_reg_count = len(wc_ta.registers)
            wc_ta.registers += [RegisterDefinition(0, 0) for i in range(len(regs), 128)]

            wc_ta.set_addr() # Update inner structure addresses

            wc_ta.registers_addr = wc_ta.registers[0]._addr
            wc_ta.register_count = ta_reg_count
            wc_ta.registers_length = ta_reg_count * 12

        wc_ta.unk_pad = 0

        if True:
            wc_ta.struct_3 = StartTACmdStruct3()
            wc_ta.struct_3.unk_480 = [0x0, 0x0, 0x0, 0x0, 0x0, 0x0] # fixed
            wc_ta.struct_3.unk_498 = 0x0 # fixed
            wc_ta.struct_3.unk_4a0 = 0x0 # fixed
            wc_ta.struct_3.iogpu_deflake_1 = deflake_1._addr
            wc_ta.struct_3.unk_4ac = 0x0 # fixed
            wc_ta.struct_3.unk_4b0 = 0x0 # fixed
            wc_ta.struct_3.unk_4b8 = 0x0 # fixed
            wc_ta.struct_3.unk_4bc = 0x0 # fixed
            wc_ta.struct_3.unk_4c4_padding = bytes(0x48)
            wc_ta.struct_3.unk_50c = 0x0 # fixed
            wc_ta.struct_3.unk_510 = 0x0 # fixed
            wc_ta.struct_3.unk_518 = 0x0 # fixed
            wc_ta.struct_3.unk_520 = 0x0 # fixed
            wc_ta.struct_3.unk_528 = 0x0 # fixed
            wc_ta.struct_3.unk_52c = 0x0 # fixed
            wc_ta.struct_3.unk_530 = 0x0 # fixed
            wc_ta.struct_3.encoder_id = cmdbuf.encoder_id
            wc_ta.struct_3.unk_538 = 0x0 # fixed
            wc_ta.struct_3.unk_53c = 0xffffffff
            wc_ta.struct_3.sampler_array = wc_3d.struct_6.sampler_array
            wc_ta.struct_3.sampler_count = 0x0
            wc_ta.struct_3.sampler_max = 0x0
            wc_ta.struct_3.unk_550 = [
                0x0, 0x0, # fixed
                0x0, # memoryless_rts_used
                0x0, 0x0, 0x0] # fixed
            wc_ta.struct_3.stamp1 = self.stamp_ta1
            wc_ta.struct_3.stamp2 = self.stamp_ta2
            wc_ta.struct_3.stamp_value = self.stamp_value_ta
            wc_ta.struct_3.ev_ta = ev_ta.id
            wc_ta.struct_3.evctl_index = self.ev_idx
            wc_ta.struct_3.unk_584 = 0x0 # flush_stamps
            wc_ta.struct_3.uuid2 = uuid_ta
            wc_ta.struct_3.queue_cmd_count = 0
            wc_ta.struct_3.unk_590 = int(unk1)

        wc_ta.set_addr() # Update inner structure addresses
        #print("wc_ta", wc_ta)

        ms = GPUMicroSequence(agx)
        if self.mshook_ta:
            self.mshook_ta(self, work, ms)

        start_ta = StartTACmd()
        if not use_registers:
            start_ta.tiling_params_addr = wc_ta.tiling_params._addr
            start_ta.struct2_addr = wc_ta.struct_2._addr # len 0x120
            start_ta.registers_addr = 0
        else:
            start_ta.tiling_params_addr = 0
            start_ta.struct2_addr = 0
            start_ta.registers_addr = wc_ta.registers[0]._addr
        start_ta.buffer_mgr = self.buffer_mgr.info
        start_ta.buf_thing = buf_desc
        start_ta.stats_ptr = agx.initdata.regionB.stats_ta.stats._addr
        start_ta.cmdqueue_ptr = self.wq_ta.info._addr
        start_ta.context_id = self.ctx_id
        start_ta.unk_38 = 1
        start_ta.submission_id = self.event_control.submission_id
        start_ta.buffer_mgr_slot = self.buffer_mgr_slot
        start_ta.unk_48 = 0#1 #0
        start_ta.unk_50 = 0
        start_ta.struct3_addr = wc_ta.struct_3._addr

        start_ta.unkptr_5c = wc_ta.unk_594._addr
        start_ta.unk_64 = 0x0 # fixed
        start_ta.unk_68 = int(unk1) # sometimes 1?
        start_ta.uuid = uuid_ta
        start_ta.unk_70 = 0x0 # fixed
        start_ta.unk_74 = [ # fixed
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
        ]
        start_ta.unk_15c = 0x0 # fixed
        start_ta.unk_160 = 0x0 # fixed
        start_ta.unk_168 = 0x0 # fixed
        start_ta.unk_16c = 0x0 # fixed
        start_ta.unk_170 = 0x0 # fixed
        start_ta.unk_178 = 0 if clustering else 1
        start_ta.counter = wc_ta.counter
        start_ta.unkptr_180 = self.event_control.unk_buf._addr

        start_ta_offset = ms.append(start_ta)

        ts1 = TimestampCmd()
        ts1.unk_1 = 0x0
        ts1.unk_2 = 0x0
        ts1.unk_3 = 0x80
        ts1.ts0_addr = wc_ta.ts1._addr
        ts1.ts_pointers_addr = wc_ta.ts_pointers._addr
        ts1.unk_addr = 0xdeadbeef
        ts1.cmdqueue_ptr = self.wq_ta.info._addr
        ts1.user_ts_pointers_addr = wc_ta.user_ts_pointers._addr
        if Ver.check("V >= V13_0B4"):
            ts1.unk_ts_addr = wc_ta.unk_ts._addr
        ts1.uuid = uuid_ta
        ts1.unk_30 = 0x0
        ms.append(ts1)

        if Ver.check("G >= G14X"):
            ms.append(Wait2Cmd())
        else:
            ms.append(WaitForInterruptCmd(1, 0, 0))

        ts2 = TimestampCmd()
        ts2.unk_1 = 0x0
        ts2.unk_2 = 0x0
        ts2.unk_3 = 0x0
        ts2.ts0_addr = wc_ta.ts1._addr
        ts2.ts_pointers_addr = wc_ta.ts_pointers._addr
        ts2.unk_addr = 0xdeadbeef
        ts2.cmdqueue_ptr = self.wq_ta.info._addr
        ts2.user_ts_pointers_addr = wc_ta.user_ts_pointers._addr
        if Ver.check("V >= V13_0B4"):
            ts2.unk_ts_addr = wc_ta.unk_ts._addr
        ts2.uuid = uuid_ta
        ts2.unk_30 = 0x0
        ms.append(ts2)

        finish_ta = FinalizeTACmd()
        finish_ta.buf_thing = buf_desc
        finish_ta.buffer_mgr = self.buffer_mgr.info
        finish_ta.stats_ptr = agx.initdata.regionB.stats_ta.stats._addr
        finish_ta.cmdqueue_ptr = self.wq_ta.info._addr
        finish_ta.context_id = self.ctx_id
        finish_ta.unk_28 = 0x0 # fixed
        finish_ta.struct3_addr = wc_ta.struct_3._addr
        finish_ta.unk_34 = 0x0 # fixed
        finish_ta.uuid = uuid_ta
        finish_ta.stamp = self.stamp_ta2
        finish_ta.stamp_value = self.stamp_value_ta
        finish_ta.unk_48 = 0x0 # fixed
        finish_ta.unk_50 = 0x0 # fixed
        finish_ta.unk_54 = 0x0 # fixed
        finish_ta.unk_58 = 0x0 # fixed
        finish_ta.unk_60 = 0x0 # fixed
        finish_ta.unk_64 = 0x0 # fixed
        finish_ta.unk_68 = 0x0 # fixed
        finish_ta.unk_6c_g14 = 0 # fixed
        finish_ta.restart_branch_offset = start_ta_offset - ms.off
        finish_ta.unk_70 = 0x0 # fixed
        finish_ta.unk_74 = bytes(0x10) # Ventura
        ms.append(finish_ta)

        ms.finalize()

        print(wc_ta)
        print(hex(wc_ta.struct_3._addr))
        print(hex(finish_ta.struct3_addr))
        print(hex(wc_ta._addr))
        ms.dump()

        work.add(ms.obj)

        wc_ta.unkptr_45c = self.tpc._addr
        wc_ta.tvb_size = tpc_size
        wc_ta.microsequence_ptr = ms.obj._addr
        wc_ta.microsequence_size = ms.size
        wc_ta.ev_3d = ev_3d.id
        wc_ta.stamp_value = self.stamp_value_ta

        print(wc_ta)
        self.wq_ta.submit(wc_ta)

        self.agx.log("Submit done")

        #self.ev_idx = (self.ev_idx + 1) % 4

        return work

    def run(self):
        ##### Run queues
        self.agx.log("Run queues")
        self.agx.ch.queue[self.queue].q_3D.run(self.wq_3d, self.ev_3d.id)
        self.agx.ch.queue[self.queue].q_TA.run(self.wq_ta, self.ev_ta.id)
        self.agx.log("Run done")

    def wait(self):
        self.agx.log("Waiting...")
        work = self.work[-1]

        ##### Wait for work completion
        while True:
            self.stamp_3d1.pull()
            self.stamp_ta1.pull()
            if (self.stamp_3d1.value >= self.stamp_value_3d and
                self.stamp_ta1.value >= self.stamp_value_ta):
                break
            self.agx.wait_for_events(timeout=2.0)

        if not self.ev_3d.fired:
            self.agx.log("3D event didn't fire")

        self.agx.log(f"Event {self.ev_3d.id} fired")
        #print("Stamps:")
        #print(self.stamp_ta1.pull())
        #print(self.stamp_ta2.pull())
        #print(self.stamp_3d1.pull())
        #print(self.stamp_3d2.pull())

        #print("WCs:")
        #print(work.wc_3d.pull())
        #print(work.wc_ta.pull())

        #if work.fb is not None and work.width and work.height:
        if work.fb is not None and work.width and work.height and work.width == 1920:
            agx = self.agx
            self.agx.log(f"Render {work.width}x{work.height} @ {work.fb:#x}")
            base, obj = self.agx.find_object(work.fb, self.ctx_id)

            #unswizzle(agx, obj._paddr, work.width, work.height, 4, "fb.bin", grid=False)
            #open("fb.bin", "wb").write(self.agx.u.iface.readmem(obj._paddr, work.width*work.height*4))
            #os.system(f"convert -size {work.width}x{work.height} -depth 8 rgba:fb.bin -alpha off frame{self.frames}.png")
            self.agx.p.fb_blit(0, 0, work.width, work.height, obj._paddr, work.width, PIX_FMT.XBGR)

        if False: #work.depth is not None:
            base, obj = self.agx.find_object(work.depth, self.ctx_id)

            width = align_up(work.width, 64)
            height = align_up(work.height, 64)

            obj.pull()
            chexdump(obj.val)

            unswizzle(self.agx, obj._paddr, work.width, work.height, 4, "depth.bin", grid=False)
            os.system(f"convert -size {work.width}x{work.height} -depth 8 rgba:depth.bin -alpha off depth.png")

        for i in self.work:
            i.free()

        self.work = []



# SPDX-License-Identifier: MIT
import itertools, fnmatch, sys
from construct import *
import sys

from .utils import AddrLookup, FourCC, SafeGreedyRange

__all__ = ["load_adt"]

ADTPropertyStruct = Struct(
    "name" / PaddedString(32, "ascii"),
    "size" / Int32ul,
    "value" / Bytes(this.size & 0x7fffffff)
)

ADTNodeStruct = Struct(
    "property_count" / Int32ul,
    "child_count" / Int32ul,
    "properties" / Array(this.property_count, Aligned(4, ADTPropertyStruct)),
    "children" / Array(this.child_count, LazyBound(lambda: ADTNodeStruct))
)

ADTStringList = SafeGreedyRange(CString("ascii"))

ADT2Tuple = Array(2, Hex(Int64ul))
ADT3Tuple = Array(3, Hex(Int64ul))

Function = Struct(
    "phandle" / Int32ul,
    "name" / FourCC,
    "args" / SafeGreedyRange(Int32ul),
)

STD_PROPERTIES = {
    "cpu-impl-reg": ADT2Tuple,
    "name": CString("ascii"),
    "compatible": ADTStringList,
    "model": CString("ascii"),
    "#size-cells": Int32ul,
    "#address-cells": Int32ul,
    "clock-ids": SafeGreedyRange(Int32ul),
    "clock-gates": SafeGreedyRange(Int32ul),
    "power-gates": SafeGreedyRange(Int32ul),
}

PMAPIORanges = SafeGreedyRange(Struct(
    "addr" / Hex(Int64ul),
    "size" / Hex(Int64ul),
    "flags" / Hex(Int32ul),
    "name" / FourCC,
))

PMGRPSRegs = SafeGreedyRange(Struct(
    "reg" / Int32ul,
    "offset" / Hex(Int32ul),
    "mask" / Hex(Int32ul),
))

PMGRPerfRegs = SafeGreedyRange(Struct(
    "reg" / Int32ul,
    "offset" / Hex(Int32ul),
    "size" / Hex(Int32ul),
    "unk" / Int32ul,
))

PMGRPWRGateRegs = SafeGreedyRange(Struct(
    "reg" / Int32ul,
    "offset" / Hex(Int32ul),
    "mask" / Hex(Int32ul),
    "unk" / Hex(Int32ul),
))

PMGRDeviceFlags = BitStruct(
    "b7" / Flag,
    "b6" / Flag,
    "perf" / Flag,
    "no_ps" / Flag,
    "critical" / Flag,
    "b2" / Flag,
    "notify_pmp" / Flag,
    "on" / Flag,
)

PMGRDevices = SafeGreedyRange(Struct(
    "flags" / PMGRDeviceFlags,
    "unk1_0" / Int8ul,
    "unk1_1" / Int8ul,
    "unk1_2" / Int8ul,
    "parents" / Array(2, Int16ul),
    "perf_idx" / Int8ul,
    "perf_block" / Int8ul,
    "psidx" / Int8ul,
    "psreg" / Int8ul,
    "unk2_0" / Int16ul,
    "pd" / Int8ul,
    "ps_cfg16" / Int8ul,
    Const(0, Int32ul),
    Const(0, Int32ul),
    "unk2_3" / Int16ul,
    "id" / Int16ul,
    "unk3" / Int32ul,
    "name" / PaddedString(16, "ascii")
))

PMGRClocks = SafeGreedyRange(Struct(
    "perf_idx" / Int8ul,
    "perf_block" / Int8ul,
    "unk" / Int8ul,
    "id" / Int8ul,
    Const(0, Int32ul),
    "name" / PaddedString(16, "ascii"),
))

PMGRPowerDomains = SafeGreedyRange(Struct(
    "unk" / Int8ul,
    "perf_idx" / Int8ul,
    "perf_block" / Int8ul,
    "id" / Int8ul,
    "flags" / Int32ul,
    "name" / PaddedString(16, "ascii"),
))

PMGRDeviceBridges = SafeGreedyRange(Struct(
    "idx" / Int32ub,
    "subdevs" / HexDump(Bytes(0x48)),
))

PMGREvents = SafeGreedyRange(Struct(
    "unk1" / Int8ul,
    "unk2" / Int8ul,
    "unk3" / Int8ul,
    "id" / Int8ul,
    "perf2_idx" / Int8ul,
    "perf2_block" / Int8ul,
    "perf_idx" / Int8ul,
    "perf_block" / Int8ul,
    "name" / PaddedString(16, "ascii"),
))

GPUPerfState = Struct(
    "freq" / Int32ul,
    "volt" / Int32ul,
)

GPUPerfState64 = Struct(
    "volt" / Int64ul,
    "freq" / Int64ul,
)

GPUPerfStates64 = Struct(
    "dies" / Int64ul,
    "count" / Int64ul,
    "states" / Array(this.dies, Array(this.count, GPUPerfState64)),
    "min_sram_volt" / Array(this.dies, Int64ul),
)

SpeakerConfig = Struct(
    "rx_slot" / Int8ul,
    "amp_gain" / Int8ul,
    "vsense_slot" / Int8ul,
    "isense_slot" / Int8ul,
)

DCBlockerConfig = Struct(
    "dc_blk0" / Hex(Int8ul),
    "dc_blk1" / Hex(Int8ul),
    "pad" / Hex(Int16ul),
)

Coef = ExprAdapter(Int32ul,
                   lambda x, ctx: (x - ((x & 0x1000000) << 1)) / 65536,
                   lambda x, ctx: int(round(x * 65536)) & 0x1ffffff)

MTRPolynomFuseAGX = GreedyRange(Struct(
    "id" / Int32ul,
    "data" / Prefixed(Int32ul, GreedyRange(Coef)),
))

SpeakerThieleSmall = Struct(
    "unk0" / Int16ul,
    "unk1" / Int16ul,
    "speakers" / GreedyRange(Struct(
        "pad0" / Hex(Int32ul),
        "r_mohm" / Int16ul,
        "temp" / Int16ul,
        "pad1" / Hex(Int32ul),
        "pad2" / Hex(Int32ul),
        "pad3" / Hex(Int16ul),
        "name" / FourCC,
    )),
    "checksum" / Hex(Int16ul),
)

TunableGlobal = Struct(
    "reg_idx" / Hex(Int32ul),
    "offset" / Hex(Int32ul),
    "mask" / Hex(Int32ul),
    "value" / Hex(Int32ul),
)

TunableLocal = Struct(
    "offset" / Hex(Int32ul),
    "size" / Hex(Int32ul),
    "mask" / Hex(Int64ul),
    "value" / Hex(Int64ul),
)

DAPFT8110 = Struct(
    "start" / Hex(Int64ul),
    "end" / Hex(Int64ul),
    "r20" / Hex(Int32ul),
    "unk1" / Hex(Int32ul),
    "r4" / Hex(Int32ul),
    "unk2" / Array(5, Hex(Int32ul)),
    "unk3" / Hex(Int8ul),
    "r0h" / Hex(Int8ul),
    "r0l" / Hex(Int8ul),
    "unk4" / Hex(Int8ul),
)

DAPFT8110B = Struct(
    "start" / Hex(Int64ul),
    "end" / Hex(Int64ul),
    "r20" / Hex(Int32ul),
    "unk1" / Hex(Int32ul),
    "r4" / Hex(Int32ul),
    "unk2" / Array(5, Hex(Int32ul)),
    "unk3" / Hex(Int8ul),
    "r0h" / Hex(Int8ul),
    "r0l" / Hex(Int8ul),
    "unk4" / Hex(Int8ul),
    "pad" / Hex(Int32ul),
)

DAPFT8110C = Struct(
    "start" / Hex(Int64ul),
    "end" / Hex(Int64ul),
    "r20" / Hex(Int32ul),
    "unk1" / Hex(Int32ul),
    "r4" / Hex(Int32ul),
    "unk2" / Array(5, Hex(Int32ul)),
    "unk3" / Hex(Int8ul),
    "r0h" / Hex(Int8ul),
    "r0l" / Hex(Int8ul),
    "unk4" / Hex(Int8ul),
    "pad" / Array(3, Hex(Int8ul)),
)

AOPRatios = Struct(
    "r0" / Int8ul,
    "r1" / Int8ul,
    "r2" / Int8ul,
    Const(0, Int8ul),
)

DEV_PROPERTIES = {
    "pmgr": {
        "*": {
            "clusters": SafeGreedyRange(Int32ul),
            "devices": PMGRDevices,
            "ps-regs": PMGRPSRegs,
            "perf-regs": PMGRPerfRegs,
            "pwrgate-regs": PMGRPWRGateRegs,
            "power-domains": PMGRPowerDomains,
            "clocks": PMGRClocks,
            "device-bridges": PMGRDeviceBridges,
            "voltage-states*": SafeGreedyRange(Int32ul),
            "events": PMGREvents,
            "mtr-polynom-fuse-agx": MTRPolynomFuseAGX,
        }
    },
    "clpc": {
        "*": {
            "events": SafeGreedyRange(Int32ul),
            "devices": SafeGreedyRange(Int32ul),
        }
    },
    "soc-tuner": {
        "*": {
            "device-set-*": SafeGreedyRange(Int32ul),
            "mcc-configs": SafeGreedyRange(Int32ul),
        }
    },
    "mcc": {
        "*": {
            "dramcfg-data": SafeGreedyRange(Int32ul),
            "config-data": SafeGreedyRange(Int32ul),
        }
    },
    "*pmu*": {
        "*": {
            "info-*name*": CString("ascii"),
            "info-*": SafeGreedyRange(Hex(Int32ul)),
        },
    },
    "stockholm-spmi": {
        "*": {
            "required-functions": ADTStringList,
        },
    },
    "sgx": {
        "*": {
            "perf-states*": SafeGreedyRange(GPUPerfState),
            "afr-perf-states": GPUPerfStates64,
            "cs-perf-states": GPUPerfStates64,
            "*-kp": Float32l,
            "*-kp-1": Float32l,
            "*-ki": Float32l,
            "*-ki-*": Float32l,
            "*-gain*": Float32l,
            "*-scale*": Float32l,
        }
    },
    "arm-io": {
        "*": {
            "clock-frequencies": SafeGreedyRange(Int32ul),
            "clock-frequencies-regs": SafeGreedyRange(Hex(Int64ul)),
            "clock-frequencies-nclk": SafeGreedyRange(Int32ul),
        },
    },
    "defaults": {
        "*": {
            "pmap-io-ranges": PMAPIORanges,
        }
    },
    "audio-*": {
        "*": {
            "speaker-config": SafeGreedyRange(SpeakerConfig),
            "amp-dcblocker-config": DCBlockerConfig,
        },
    },
    "*aop-audio*": {
        "*": {
            "clockSource": FourCC,
            "identifier": FourCC,
            "ratios": AOPRatios,
            "filterLengths": Hex(Int32ul),
        },
    },
    "*alc?/audio-leap-mic*": {
        "*": {
            "audio-stream-formatter": FourCC,
        }
    },
    "audio": {
        "*": {
            "speaker-thiele-small": SpeakerThieleSmall,
        },
    },
    "apcie*": {
        "*": {
            "apcie-*-tunables": GreedyRange(TunableLocal),
        }
    },
}

def parse_prop(node, path, node_name, name, v, is_template=False):
    t = None

    if is_template:
        t = CString("ascii")
        v = Sequence(t, Terminated).parse(v)[0]
        return t, v

    dev_props = None
    for k, pt in DEV_PROPERTIES.items():
        if fnmatch.fnmatch(path, k):
            dev_props = pt
            break

    if not dev_props:
        for k, pt in DEV_PROPERTIES.items():
            if fnmatch.fnmatch(node_name, k):
                dev_props = pt
                break

    possible_match = False
    if dev_props:
        for compat_match, cprops in dev_props.items():
            for k, pt in cprops.items():
                if fnmatch.fnmatch(name, k):
                    possible_match = True
                    break

    if possible_match:
        try:
            compat = node.compatible[0]
        except AttributeError:
            compat = ""

        for compat_match, cprops in dev_props.items():
            if fnmatch.fnmatch(compat, compat_match):
                for k, pt in cprops.items():
                    if fnmatch.fnmatch(name, k):
                        t = pt
                        break
                else:
                    continue
                break

    if v == b'' or v is None:
        return None, None

    if name.startswith("function-"):
        if len(v) == 4:
            t = FourCC
        else:
            t = Function

    if name == "reg" and path != "/device-tree/memory":
        n = node._parent
        while n is not None and n._parent is not None:
            if "ranges" not in n._properties:
                break
            n = n._parent
        else:
            rs = node._reg_struct
            if len(v) % rs.sizeof() == 0:
                t = SafeGreedyRange(rs)

    elif name == "ranges":
        try:
            ac, sc = node.address_cells, node.size_cells
        except AttributeError:
            return None, v
        pac, _ = node._parent.address_cells, node._parent.size_cells
        at = Hex(Int64ul) if ac == 2 else Array(ac, Hex(Int32ul))
        pat = Hex(Int64ul) if pac == 2 else Array(pac, Hex(Int32ul))
        st = Hex(Int64ul) if sc == 2 else Array(sc, Hex(Int32ul))
        t = SafeGreedyRange(Struct("bus_addr" / at, "parent_addr" / pat, "size" / st))

    elif name.startswith("dapf-instance-"):
        try:
            flags = node.dart_options
        except AttributeError:
            return None, v
        if flags & 0x40:
            if len(v) % DAPFT8110B.sizeof() == 0:
                if len(v) % DAPFT8110C.sizeof() != 0:
                    t = GreedyRange(DAPFT8110B)
            else:
                t = GreedyRange(DAPFT8110C)
        else:
            t = GreedyRange(DAPFT8110)

    elif name == "interrupts":
        # parse "interrupts" as Array of Int32ul, wrong for nodes whose
        # "interrupt-parent" has "interrupt-cells" = 2
        # parsing this correctly would require a second pass
        t = Array(len(v) // 4, Int32ul)

    if t is not None:
        v = Sequence(t, Terminated).parse(v)[0]
        return t, v

    if name in STD_PROPERTIES:
        t = STD_PROPERTIES[name]
    elif v and v[-1] == 0 and all(0x20 <= i <= 0x7e for i in v[:-1]):
        t = CString("ascii")
    elif len(v) == 4:
        t = Int32ul
    elif len(v) == 8:
        t = Int64ul
    elif len(v) == 16 and all(v[i] == 0 for i in (6, 7, 14, 15)):
        t = ADT2Tuple

    if t is not None:
        try:
            v = Sequence(t, Terminated).parse(v)[0]
        except:
            print("Failed to parse:", path, name, v.hex())
            raise

    return t, v

def build_prop(path, name, v, t=None):
    if v is None:
        return b''
    if t is not None:
        return t.build(v)

    if isinstance(v, bytes):
        return v

    if name in STD_PROPERTIES:
        t = STD_PROPERTIES[name]
    elif isinstance(v, str):
        t = CString("ascii")
    elif isinstance(v, int):
        if v > 0xffffffff:
            t = Int64ul
        else:
            t = Int32ul
    elif isinstance(v, float):
        t = Float32l
    elif isinstance(v, tuple) and all(isinstance(i, int) for i in v):
        t = Array(len(v), Int32ul)

    return t.build(v)

class ADTNode:
    def __init__(self, val=None, path="/", parent=None):
        self._children = []
        self._properties = {}
        self._types = {}
        self._parent_path = path
        self._parent = parent

        if val is not None:
            for p in val.properties:
                if p.name == "name":
                    _name = p.value.decode("ascii").rstrip("\0")
                    break
            else:
                raise ValueError(f"Node in {path} has no name!")

            path = self._parent_path + _name

            raw = {}
            for p in val.properties:
                raw[p.name] = p.value
                is_template = bool(p.size & 0x80000000)
                try:
                    t, v = parse_prop(self, path, _name, p.name, p.value, is_template)
                    self._types[p.name] = t, is_template
                    self._properties[p.name] = v
                except Exception as e:
                    print(f"Exception parsing {path}.{p.name} value {p.value.hex()}:", file=sys.stderr)
                    raise

            # Second pass
            for k, (t, is_template) in self._types.items():
                if t is None:
                    t, v = parse_prop(self, path, _name, k, self._properties[k], is_template)
                    self._types[k] = t, is_template
                    self._properties[k] = v
                    assert build_prop(self._path, k, v, t=t) == raw[k]

            for c in val.children:
                node = ADTNode(c, f"{self._path}/", parent=self)
                self._children.append(node)

    @property
    def _path(self):
        return self._parent_path + self.name

    def __getitem__(self, item):
        if isinstance(item, str):
            while item.startswith("/"):
                item = item[1:]
            if "/" in item:
                a, b = item.split("/", 1)
                return self[a][b]
            for i in self._children:
                if i.name == item:
                    return i
            raise KeyError(f"Child node '{item}' not found")
        return self._children[item]

    def __setitem__(self, item, value):
        if isinstance(item, str):
            while item.startswith("/"):
                item = item[1:]
            if "/" in item:
                a, b = item.split("/", 1)
                self[a][b] = value
                return
            for i, c in enumerate(self._children):
                if c.name == item:
                    self._children[i] = value
                    break
            else:
                self._children.append(value)
        else:
            self._children[item] = value

    def __delitem__(self, item):
        if isinstance(item, str):
            while item.startswith("/"):
                item = item[1:]
            if "/" in item:
                a, b = item.split("/", 1)
                del self[a][b]
                return
            for i, c in enumerate(self._children):
                if c.name == item:
                    del self._children[i]
                    return
            raise KeyError(f"Child node '{item}' not found")

        del self._children[item]

    def __contains__(self, item):
        if isinstance(item, str):
            while item.startswith("/"):
                item = item[1:]
            if "/" in item:
                a, b = item.split("/", 1)
                return b in self[a]
            for c in self._children:
                if c.name == item:
                    return True
            return False

        return item in self._children

    def __getattr__(self, attr):
        attr = attr.replace("_", "-")
        attr = attr.replace("--", "_")
        if attr in self._properties:
            return self._properties[attr]
        raise AttributeError(attr)

    def __setattr__(self, attr, value):
        if attr[0] == "_":
            self.__dict__[attr] = value
            return
        attr = attr.replace("_", "-")
        attr = attr.replace("--", "_")
        self._properties[attr] = value

    def __delattr__(self, attr):
        if attr[0] == "_":
            del self.__dict__[attr]
            return
        attr = attr.replace("_", "-")
        attr = attr.replace("--", "_")
        del self._properties[attr]

    def getprop(self, name, default=None):
        return self._properties.get(name, default)

    @property
    def address_cells(self):
        try:
            return self._properties["#address-cells"]
        except KeyError:
            raise AttributeError("#address-cells")

    @property
    def size_cells(self):
        try:
            return self._properties["#size-cells"]
        except KeyError:
            raise AttributeError("#size-cells")

    @property
    def interrupt_cells(self):
        try:
            return self._properties["#interrupt-cells"]
        except KeyError:
            raise AttributeError("#interrupt-cells")

    def _fmt_prop(self, k, v):
        t, is_template = self._types.get(k, (None, False))
        if is_template:
            return f"<< {v} >>"
        elif isinstance(v, ListContainer):
            return f"[{', '.join(self._fmt_prop(k, i) for i in v)}]"
        elif isinstance(v, bytes):
            if all(i == 0 for i in v):
                return f"zeroes({len(v):#x})"
            else:
                return v.hex()
        elif k.startswith("function-"):
            if isinstance(v, str):
                return f"{v}()"
            elif v is None:
                return f"None"
            else:
                args = []
                for arg in v.args:
                    b = arg.to_bytes(4, "big")
                    is_ascii = all(0x20 <= c <= 0x7e for c in b)
                    args.append(f"{arg:#x}" if not is_ascii else f"'{b.decode('ascii')}'")
                return f"{v.phandle}:{v.name}({', '.join(args)})"
            name.startswith("function-")
        else:
            return str(v)

    def __str__(self, t=""):
        return "\n".join([
            t + f"{self.name} {{",
            *(t + f"    {k} = {self._fmt_prop(k, v)}" for k, v in self._properties.items() if k != "name"),
            "",
            *(i.__str__(t + "    ") for i in self._children),
            t + "}"
        ])

    def __repr__(self):
        return f"<ADTNode {self.name}>"

    def __iter__(self):
        return iter(self._children)

    @property
    def _reg_struct(self):
        ac, sc = self._parent.address_cells, self._parent.size_cells
        return Struct(
            "addr" / Hex(Int64ul) if ac == 2 else Array(ac, Hex(Int32ul)),
            "size" / Hex(Int64ul) if sc == 2 else Array(sc, Hex(Int32ul))
        )

    def get_reg(self, idx):
        reg = self.reg[idx]
        addr = reg.addr
        size = reg.size

        return self._parent.translate(addr), size

    def translate(self, addr):
        node = self
        while node is not None:
            if "ranges" not in node._properties:
                break
            for r in node.ranges:
                ba = r.bus_addr
                # PCIe special case, because Apple really broke
                # the spec here with their little endian antics
                if isinstance(ba, list) and len(ba) == 3:
                    ba = (ba[0] << 64) | (ba[2] << 32) | ba[1]
                if ba <= addr < (ba + r.size):
                    addr = addr - ba + r.parent_addr
                    break
            node = node._parent

        return addr

    def to_bus_addr(self, addr):
        node = self._parent

        descend = []
        while node is not None:
            if "ranges" not in node._properties:
                break
            descend.append(node)
            node = node._parent

        for node in reversed(descend):
            for r in node.ranges:
                if r.parent_addr <= addr < (r.parent_addr + r.size):
                    addr = addr - r.parent_addr + r.bus_addr
                    break
        return addr

    def tostruct(self):
        properties = []
        for k,v in itertools.chain(self._properties.items()):
            t, is_template = self._types.get(k, (None, False))
            value = build_prop(self._path, k, v, t=t)
            properties.append({
                "name": k,
                "size": len(value) | (0x80000000 if is_template else 0),
                "value": value
            })

        data = {
            "property_count": len(self._properties),
            "child_count": len(self._children),
            "properties": properties,
            "children": [c.tostruct() for c in self._children]
        }
        return data

    def build(self):
        return ADTNodeStruct.build(self.tostruct())

    def walk_tree(self):
        yield self
        for child in self:
            yield from child

    def build_addr_lookup(self):
        lookup = AddrLookup()
        for node in self.walk_tree():
            reg = getattr(node, 'reg', None)
            if not isinstance(reg, list):
                continue

            for index in range(len(reg)):
                try:
                    addr, size = node.get_reg(index)
                except AttributeError:
                    continue
                if size == 0:
                    continue
                lookup.add(range(addr, addr + size), node.name + f"[{index}]")

        return lookup

    def create_node(self, name):
        while name.startswith("/"):
            name = name[1:]
        if "/" in name:
            a, b = name.split("/", 1)
            return self[a].create_node(b)

        node = ADTNode(path=self._path + "/", parent=self)
        node.name = name
        node._types["reg"] = (SafeGreedyRange(node._reg_struct), False)
        self[name] = node
        return node

def load_adt(data):
    return ADTNode(ADTNodeStruct.parse(data))

if __name__ == "__main__":
    import sys, argparse, pathlib

    parser = argparse.ArgumentParser(description='ADT test for m1n1')
    parser.add_argument('input', type=pathlib.Path)
    parser.add_argument('output', nargs='?', type=pathlib.Path)
    parser.add_argument('-r', '--retrieve', help='retrieve and store the adt from m1n1', action='store_true')
    parser.add_argument('-a', '--dump-addr', help='dump address lookup table', action='store_true')
    args = parser.parse_args()

    if args.retrieve:
        if args.input.exists():
            print('Error "{}" exists!'.format(args.input))
            sys.exit()

        from .setup import *
        adt_data = u.get_adt()
        args.input.write_bytes(adt_data)
    else:
        adt_data = args.input.read_bytes()

    adt = load_adt(adt_data)
    print(adt)
    new_data = adt.build()
    if args.output is not None:
        args.output.write_bytes(new_data)
    assert new_data == adt_data[:len(new_data)]
    assert adt_data[len(new_data):] == bytes(len(adt_data) - len(new_data))

    if args.dump_addr:
        print("Address lookup table:")
        print(adt.build_addr_lookup())
